# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
This script prints out all teams existing in the space of the given
organization
"""

from argparse import ArgumentParser, Namespace

from rich.console import Console
from rich.table import Table

from pontos.github.api import GitHubAsyncRESTApi


def add_script_arguments(parser: ArgumentParser) -> None:
    parser.add_argument("organization")


async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:
    table = Table()
    table.add_column("Name")
    table.add_column("Description")
    table.add_column("URL")
    table.add_column("Privacy")

    count = 0
    async for team in api.teams.get_all(args.organization):
        table.add_row(
            team.name,
            team.description,
            f"[link={team.html_url}]{team.html_url}[/link]",
            team.privacy.value,
        )
        count += 1

    console = Console()
    console.print(table)

    print(f"{count} teams.")
    return 0



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from .core import ActionIO, ActionOutput, Console
from .env import GitHubEnvironment
from .errors import GitHubActionsError
from .event import (
    GitHubEvent,
    GitHubPullRequestEvent,
    Label,
    PullRequestState,
    Ref,
)
from .main import main

__all__ = (
    "GitHubActionsError",
    "ActionIO",
    "ActionOutput",
    "Console",
    "GitHubEnvironment",
    "GitHubEvent",
    "Label",
    "Ref",
    "PullRequestState",
    "GitHubPullRequestEvent",
    "main",
)



# SPDX-FileCopyrightText: 2021-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import contextlib
import os
import subprocess
import unittest
from pathlib import Path
from typing import Generator
from unittest.mock import MagicMock, patch

from pontos.git import ConfigScope, Git, GitError
from pontos.release.helper import (
    ReleaseType,
    find_signing_key,
    get_git_repository_name,
    get_next_release_version,
    repository_split,
)
from pontos.testing import temp_git_repository
from pontos.version import VersionError
from pontos.version.schemes import SemanticVersioningScheme


@contextlib.contextmanager
def init_test_git_repo() -> Generator[Path, None, None]:
    with temp_git_repository() as repo_path:
        git = Git()
        git.add_remote("foo", "https://foo.bar/bla.git")
        git.add_remote("origin", "https://foo.bar/testrepo.git")
        yield repo_path


class GetGitRepositoryNameTestCase(unittest.TestCase):
    def test_get_project_name(self):
        with init_test_git_repo():
            project = get_git_repository_name(remote="foo")
            self.assertEqual(project, "bla")

            project = get_git_repository_name()
            self.assertEqual(project, "testrepo")


class FindSigningKeyTestCase(unittest.TestCase):
    def test_find_signing_key(self):
        terminal = MagicMock()

        with temp_git_repository():
            git = Git()
            git.config(
                "user.signingkey",
                "1234567890ABCEDEF1234567890ABCEDEF123456",
                scope=ConfigScope.LOCAL,
            )
            signing_key = find_signing_key(terminal)
            self.assertEqual(
                signing_key, "1234567890ABCEDEF1234567890ABCEDEF123456"
            )

    @unittest.skipUnless(os.environ.get("CI"), "only run on CI")
    def test_find_no_signing_key(self):
        terminal = MagicMock()
        saved_key = None

        git = Git()
        try:
            # save possibly set git signing key from user temporarily
            try:
                saved_key = git.config(
                    "user.signingkey", scope=ConfigScope.GLOBAL
                )
            except GitError:
                saved_key = None

            try:
                git.config("user.signingkey", "", scope=ConfigScope.GLOBAL)
            except subprocess.CalledProcessError as e:
                self.assertEqual(e.returncode, 5)

            with temp_git_repository():
                signing_key = find_signing_key(terminal)
                self.assertEqual(signing_key, "")

        finally:
            # reset the previously saved signing key ...
            if saved_key is not None:
                git.config(
                    "user.signingkey", saved_key, scope=ConfigScope.GLOBAL
                )


class GetNextReleaseVersionTestCase(unittest.TestCase):
    def test_next_major_version(self):
        calculator = SemanticVersioningScheme.calculator()
        last_version = SemanticVersioningScheme.parse_version("1.0.0")
        version = get_next_release_version(
            calculator=calculator,
            last_release_version=last_version,
            release_type=ReleaseType.MAJOR,
            release_version=None,
        )

        self.assertEqual(
            version, SemanticVersioningScheme.parse_version("2.0.0")
        )

    def test_next_minor_version(self):
        calculator = SemanticVersioningScheme.calculator()
        last_version = SemanticVersioningScheme.parse_version("1.0.0")
        version = get_next_release_version(
            calculator=calculator,
            last_release_version=last_version,
            release_type=ReleaseType.MINOR,
            release_version=None,
        )

        self.assertEqual(
            version, SemanticVersioningScheme.parse_version("1.1.0")
        )

    def test_next_patch_version(self):
        calculator = SemanticVersioningScheme.calculator()
        last_version = SemanticVersioningScheme.parse_version("1.0.0")
        version = get_next_release_version(
            calculator=calculator,
            last_release_version=last_version,
            release_type=ReleaseType.PATCH,
            release_version=None,
        )

        self.assertEqual(
            version, SemanticVersioningScheme.parse_version("1.0.1")
        )

    def test_next_calendar_version(self):
        calculator = SemanticVersioningScheme.calculator()
        last_version = SemanticVersioningScheme.parse_version("1.0.0")

        with patch.object(
            calculator,
            "next_calendar_version",
            return_value=SemanticVersioningScheme.parse_version("23.7.0"),
        ):
            version = get_next_release_version(
                calculator=calculator,
                last_release_version=last_version,
                release_type=ReleaseType.CALENDAR,
                release_version=None,
            )

        self.assertEqual(
            version, SemanticVersioningScheme.parse_version("23.7.0")
        )

    def test_next_alpha_version(self):
        calculator = SemanticVersioningScheme.calculator()
        last_version = SemanticVersioningScheme.parse_version("1.0.0")
        version = get_next_release_version(
            calculator=calculator,
            last_release_version=last_version,
            release_type=ReleaseType.ALPHA,
            release_version=None,
        )

        self.assertEqual(
            version, SemanticVersioningScheme.parse_version("1.0.1-alpha1")
        )

    def test_next_beta_version(self):
        calculator = SemanticVersioningScheme.calculator()
        last_version = SemanticVersioningScheme.parse_version("1.0.0")
        version = get_next_release_version(
            calculator=calculator,
            last_release_version=last_version,
            release_type=ReleaseType.BETA,
            release_version=None,
        )

        self.assertEqual(
            version, SemanticVersioningScheme.parse_version("1.0.1-beta1")
        )

    def test_next_release_candidate_version(self):
        calculator = SemanticVersioningScheme.calculator()
        last_version = SemanticVersioningScheme.parse_version("1.0.0")
        version = get_next_release_version(
            calculator=calculator,
            last_release_version=last_version,
            release_type=ReleaseType.RELEASE_CANDIDATE,
            release_version=None,
        )

        self.assertEqual(
            version, SemanticVersioningScheme.parse_version("1.0.1-rc1")
        )

    def test_no_release_type(self):
        calculator = SemanticVersioningScheme.calculator()
        last_version = SemanticVersioningScheme.parse_version("1.0.0")
        next_version = SemanticVersioningScheme.parse_version("1.2.3")
        version = get_next_release_version(
            calculator=calculator,
            last_release_version=last_version,
            release_type=None,
            release_version=next_version,
        )

        self.assertEqual(version, next_version)

    def test_release_type_version(self):
        calculator = SemanticVersioningScheme.calculator()
        last_version = SemanticVersioningScheme.parse_version("1.0.0")
        next_version = SemanticVersioningScheme.parse_version("1.2.3")
        version = get_next_release_version(
            calculator=calculator,
            last_release_version=last_version,
            release_type=ReleaseType.VERSION,
            release_version=next_version,
        )

        self.assertEqual(version, next_version)

    def test_no_release_type_and_release_version(self):
        calculator = SemanticVersioningScheme.calculator()
        last_version = SemanticVersioningScheme.parse_version("1.0.0")

        with self.assertRaisesRegex(
            VersionError,
            "No release version provided. Either use a different release "
            "type or provide a release version.",
        ):
            get_next_release_version(
                calculator=calculator,
                last_release_version=last_version,
                release_type=None,
                release_version=None,
            )

    def test_no_release_type_and_no_last_release_version(self):
        calculator = SemanticVersioningScheme.calculator()
        release_version = SemanticVersioningScheme.parse_version("1.0.0")

        version = get_next_release_version(
            calculator=calculator,
            last_release_version=None,
            release_type=None,
            release_version=release_version,
        )

        self.assertEqual(version, release_version)

    def test_release_type_version_and_no_last_release_version(self):
        calculator = SemanticVersioningScheme.calculator()
        release_version = SemanticVersioningScheme.parse_version("1.0.0")

        version = get_next_release_version(
            calculator=calculator,
            last_release_version=None,
            release_type=ReleaseType.VERSION,
            release_version=release_version,
        )

        self.assertEqual(version, release_version)

    def test_no_last_release_version(self):
        calculator = SemanticVersioningScheme.calculator()

        with self.assertRaisesRegex(
            VersionError,
            "No last release version found for release type alpha. Either "
            "check the project setup or set a release version explicitly.",
        ):
            get_next_release_version(
                calculator=calculator,
                last_release_version=None,
                release_type=ReleaseType.ALPHA,
                release_version=None,
            )

    def test_release_version_with_invalid_release_type(self):
        calculator = SemanticVersioningScheme.calculator()
        release_version = SemanticVersioningScheme.parse_version("1.0.0")

        with self.assertRaisesRegex(
            VersionError,
            "Invalid release type alpha when setting release version "
            "explicitly. Use release type version instead.",
        ):
            get_next_release_version(
                calculator=calculator,
                last_release_version=None,
                release_type=ReleaseType.ALPHA,
                release_version=release_version,
            )


class RepositorySplitTestCase(unittest.TestCase):
    def test_invalid_repository(self):
        with self.assertRaisesRegex(
            ValueError,
            r"Invalid repository foo/bar/baz. Format must be owner/name.",
        ):
            repository_split("foo/bar/baz")

        with self.assertRaisesRegex(
            ValueError,
            r"Invalid repository foo_bar_baz. Format must be owner/name.",
        ):
            repository_split("foo_bar_baz")

    def test_repository(self):
        space, project = repository_split("foo/bar")

        self.assertEqual(space, "foo")
        self.assertEqual(project, "bar")



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from .api import GitHubAsyncRESTApi
from .artifacts import GitHubAsyncRESTArtifacts
from .branch import GitHubAsyncRESTBranches, update_from_applied_settings
from .contents import GitHubAsyncRESTContent
from .dependabot import GitHubAsyncRESTDependabot
from .errors import GitHubApiError
from .helper import (
    DEFAULT_GITHUB_API_URL,
    DEFAULT_TIMEOUT_CONFIG,
    JSON,
    JSON_OBJECT,
)
from .labels import GitHubAsyncRESTLabels
from .organizations import GitHubAsyncRESTOrganizations
from .packages import GitHubAsyncRESTPackages
from .pull_requests import GitHubAsyncRESTPullRequests
from .release import GitHubAsyncRESTReleases
from .repositories import GitHubAsyncRESTRepositories
from .search import GitHubAsyncRESTSearch
from .tags import GitHubAsyncRESTTags
from .teams import GitHubAsyncRESTTeams
from .workflows import GitHubAsyncRESTWorkflows

__all__ = [
    "DEFAULT_TIMEOUT_CONFIG",
    "DEFAULT_GITHUB_API_URL",
    "JSON",
    "JSON_OBJECT",
    "update_from_applied_settings",
    "GitHubApiError",
    "GitHubAsyncRESTApi",
    "GitHubAsyncRESTArtifacts",
    "GitHubAsyncRESTBranches",
    "GitHubAsyncRESTContent",
    "GitHubAsyncRESTDependabot",
    "GitHubAsyncRESTLabels",
    "GitHubAsyncRESTOrganizations",
    "GitHubAsyncRESTPackages",
    "GitHubAsyncRESTPullRequests",
    "GitHubAsyncRESTReleases",
    "GitHubAsyncRESTRepositories",
    "GitHubAsyncRESTSearch",
    "GitHubAsyncRESTTags",
    "GitHubAsyncRESTTeams",
    "GitHubAsyncRESTWorkflows",
]



# SPDX-FileCopyrightText: 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import re
import urllib.parse
from dataclasses import dataclass
from typing import Any, Optional

from pontos.errors import PontosError
from pontos.models import StrEnum

__all__ = (
    "ANY",
    "NA",
    "CPEParsingError",
    "Part",
    "CPE",
)

ANY = "*"
NA = "-"


class CPEParsingError(PontosError):
    """
    An error occurred while parsing a CPE
    """


class Part(StrEnum):
    """
    Represents the possible values for a part CPE attribute
    """

    APPLICATION = "a"
    OPERATING_SYSTEM = "o"
    HARDWARE_DEVICE = "h"


def is_uri_binding(cpe: str) -> bool:
    """
    Returns True if cpe is a CPE v2.2 URI string
    """
    return cpe.startswith("cpe:/")


def is_formatted_string_binding(cpe: str) -> bool:
    """
    Returns True if cpe is a CPE v2.3 formatted string
    """
    return cpe.startswith("cpe:2.3:")


def convert_double_backslash(value: str) -> str:
    """
    Convert a double backslash into s single backslash
    """
    return re.sub("\\\\(\\W)", lambda match: match.group(1), value)


def _url_quote(value: str) -> str:
    """
    Quote value according to the pct_encode function from the spec for uri format
    """
    return urllib.parse.quote(value, safe="").lower()


def _url_unquote(value: str) -> str:
    """
    Un-quote value according to the the spec for uri format
    """
    return urllib.parse.unquote(value)


def pack_extended_attributes(
    edition: Optional[str],
    sw_edition: Optional[str],
    target_sw: Optional[str],
    target_hw: Optional[str],
    other: Optional[str],
) -> str:
    """
    Pack the extended attributes (v2.3) for an edition attribute (v2.2)
    """
    if (
        (not sw_edition or sw_edition == ANY)
        and (not target_sw or target_sw == ANY)
        and (not target_hw or target_hw == ANY)
        and (not other or other == ANY)
    ):
        if not edition or edition == ANY:
            return ""
        else:
            return edition
    else:
        return (
            f"~{'' if not edition or edition == ANY else edition}"
            f"~{'' if not sw_edition or sw_edition == ANY else sw_edition}"
            f"~{'' if not target_sw or target_sw == ANY else target_sw}"
            f"~{'' if not target_hw or target_hw == ANY else target_hw}"
            f"~{'' if not other or other == ANY else other}"
        )


def unpack_edition(edition: str) -> dict[str, Optional[str]]:
    """
    Unpack the edition attribute of v2.2 into extended attributes of v2.3
    """
    return dict(
        zip(
            [
                "edition",
                "sw_edition",
                "target_sw",
                "target_hw",
                "other",
            ],
            [None if not a else a for a in edition.split("~")[1:-1]],
        )
    )


def bind_value_for_formatted_string(value: Optional[str]) -> str:
    """
    Convert an attribute value for formatted string representation
    """
    if not value or value == ANY:
        return ANY

    value = value.replace("\\.", ".")
    value = value.replace("\\-", "-")
    value = value.replace("\\_", "_")
    return value


def _add_quoting(value: str) -> str:
    """
    Add quoting for parsing attributes from formatted string format to
    Well-Formed CPE Name Data Model (WFN)
    """
    result: list[str] = []
    index = 0
    embedded = False

    while index < len(value):
        c = value[index]
        if c.isalnum() or c in ["_"]:
            # just add character
            result.append(c)
            index += 1
            embedded = True
            continue

        if c == "\\":
            # keep escaped character
            result.append(value[index : index + 2])
            index += 2
            embedded = True
            continue

        if c == ANY:
            # An unquoted asterisk must appear at the beginning or
            # end of the string.
            if index == 0 or index == (len(value) - 1):
                result.append(c)
                index += 1
                embedded = True
                continue
            else:
                raise CPEParsingError(
                    "An unquoted asterisk must appear at the beginning or end "
                    f"of '{value}'"
                )
        if c == "?":
            # An unquoted question mark must appear at the beginning or
            # end of the string, or in a leading or trailing sequence
            if (
                (  # ? is legal at the beginning or the end
                    (index == 0) or (index == (len(value) - 1))
                )
                or (  # embedded is false, so must be preceded by ?
                    not embedded and (value[index - 1 : index] == "?")
                )
                or (  # embedded is true, so must be followed by ?
                    embedded and (value[index + 1] == "?")
                )
            ):
                result.append(c)
                index += 1
                embedded = False
                continue
            else:
                raise CPEParsingError(
                    "An unquoted question mark must appear at the beginning or "
                    f"end, or in a leading or trailing sequence '{value}'"
                )

        # all other characters must be quoted
        result.append(f"\\{c}")
        index += 1
        embedded = True

    return "".join(result)


def unbind_value_from_formatted_string(value: Optional[str]) -> Optional[str]:
    """
    Convert a formatted string representation to an attribute value for WNF
    """
    if value is None or value == ANY or value == NA:
        return value
    return _add_quoting(value)


def _transform_for_uri(value: str) -> str:
    """
    Applies transform to convert an attribute for an uri representation

    The following transformations are applied:
        - Pass alphanumeric characters thru untouched
        - Percent-encode quoted non-alphanumerics as needed
        - Unquoted special characters are mapped to their special forms
    """
    transformed = ""
    index = 0
    while index < len(value):
        c = value[index]

        # alpha numeric characters
        if c.isalnum() or c in ["_", "-", ".", "~"]:
            transformed += c
            index += 1
            continue

        # percent encoding
        if c == "\\":
            index += 1
            next = value[index]
            transformed += _url_quote(convert_double_backslash(next))
            index += 1
            continue

        # special forms
        if c == "?":
            transformed += "%01"
        elif c == "*":
            transformed += "%02"

        index += 1

    return transformed


def bind_value_for_uri(value: Optional[str]) -> str:
    """
    Convert an attribute value for uri representation
    """
    if not value or value == ANY:
        return ""
    if value == NA:
        return value
    try:
        return _transform_for_uri(value)
    except Exception as e:
        raise CPEParsingError(f"Can't bind '{value}' for URI") from e


def unbind_value_uri(value: Optional[str]) -> Optional[str]:
    """
    Convert an uri representation to an attribute value
    """
    if value is None:
        return None
    if value == "":
        return ANY
    if value == NA:
        return NA

    result = ""
    index = 0
    embedded = False
    while index < len(value):
        c = value[index]
        if c == "." or c == "-" or c == "~":
            result += f"\\{c}"
            index += 1
            embedded = True
            continue

        if c != "%":
            result += c
            index += 1
            embedded = True
            continue

        form = value[index : index + 3]
        if form == "%01":
            if (
                index == 0
                or (index == (len(value) - 3))
                or (not embedded and (value[index - 3 : index] == "%01"))
                or (
                    embedded
                    and (len(value) >= index + 6)
                    and (value[index + 3 : index + 6]) == "%01"
                )
            ):
                result += "?"
            else:
                raise CPEParsingError(
                    "A percent-encoded question mark is not found at the "
                    f"beginning or the end or embedded in sequence '{value}'"
                )
        elif form == "%02":
            if (index == 0) or (index == (len(value) - 3)):
                result += "*"
            else:
                raise CPEParsingError(
                    "Percent-encoded asterisk is no at the beginning "
                    f"or the end of '{value}'"
                )
        else:
            result += f"\\{_url_unquote(form)}"
        index += 3
        embedded = True

    return result


def unquote_attribute_value(value: Optional[str]) -> Optional[str]:
    """
    Unquote a Well-Formed CPE Name Data Model (WFN) attribute value
    """
    if not value or "\\" not in value:
        # do nothing
        return value

    index = 0
    result = ""
    while index < len(value):
        c = value[index]
        if c == "\\":
            next_c = value[index + 1]
            if next_c in ["*", "?"]:
                # keep escaped asterisks and question marks
                result += f"{c}{next_c}"
            else:
                result += next_c

            index += 2
            continue
        else:
            result += c

        index += 1

    return result


def split_cpe(cpe: str) -> list[str]:
    """
    Split a CPE into its parts
    """
    if "\\:" in cpe:
        # houston we have a problem
        # the cpe string contains an escaped colon (:)
        parts = []
        index = 0
        start_index = 0
        stripped_cpe = cpe
        while index < len(cpe):
            if index > 0 and cpe[index] == ":" and cpe[index - 1] != "\\":
                part = cpe[start_index:index]
                parts.append(part)
                start_index = index + 1
                stripped_cpe = cpe[start_index:]
            index += 1

        if stripped_cpe:
            parts.append(stripped_cpe)
    else:
        parts = cpe.split(":")

    return parts


@dataclass(frozen=True)
class CPEWellFormed:
    """
    Represents a Common Platform Enumeration (CPE) name using the Well-Formed
    CPE Name (WNF) Data Model. Attributes are quoted according to the WNF model.

    In most cases this class should not be used directly and the CPE class
    should be used instead.

    Attributes:
        part: Value should be "a" for application, "o" for operating system or
            "h" for hardware
        vendor: Person or organization that manufactured or created the product
        product: Identifies the most common and recognizable title or name of
            the product
        version: A vendor-specific alphanumeric string characterizing the
            particular release version of the product
        update: A vendor-specific alphanumeric string characterizing the
            particular update, service pack, or point release of the product
        edition: The edition attribute is considered deprecated in the 2.3
            CPE specification, and it should be assigned the logical value ANY
            except where required for backward compatibility with version 2.2 of
            the CPE specification. This attribute is referred to as the “legacy
            edition” attribute
        language: Defines the language supported in the user interface of the
            product (as language tags defined by RFC5646)
        sw_edition: Characterizes how the product is tailored to a particular
            market or class of end users. Extended attribute introduced with
            version 2.3 of the CPE specification
        target_sw: Characterizes the software computing environment within which
            the product operates. Extended attribute introduced with
            version 2.3 of the CPE specification
        hardware_sw: Characterizes the instruction set architecture (e.g., x86)
            on which the product operates. Extended attribute introduced with
            version 2.3 of the CPE specification
        other: Captures any other general descriptive or identifying information
            which is vendor- or product-specific and which does not logically
            fit in any other attribute value. Extended attribute introduced with
            version 2.3 of the CPE specification
    """

    part: Part
    vendor: str
    product: str
    version: Optional[str] = None
    update: Optional[str] = None
    edition: Optional[str] = None
    language: Optional[str] = None
    sw_edition: Optional[str] = None
    target_sw: Optional[str] = None
    target_hw: Optional[str] = None
    other: Optional[str] = None


class CPE:
    """
    Represents a Common Platform Enumeration (CPE) name

    Supports CPE specification 2.2 (uri) and 2.3 (formatted string)

    Attributes:
        part: Value should be "a" for application, "o" for operating system or
            "h" for hardware
        vendor: Person or organization that manufactured or created the product
        product: Identifies the most common and recognizable title or name of
            the product
        version: A vendor-specific alphanumeric string characterizing the
            particular release version of the product
        update: A vendor-specific alphanumeric string characterizing the
            particular update, service pack, or point release of the product
        edition: The edition attribute is considered deprecated in the 2.3
            CPE specification, and it should be assigned the logical value ANY
            except where required for backward compatibility with version 2.2 of
            the CPE specification. This attribute is referred to as the “legacy
            edition” attribute
        language: Defines the language supported in the user interface of the
            product (as language tags defined by RFC5646)
        sw_edition: Characterizes how the product is tailored to a particular
            market or class of end users. Extended attribute introduced with
            version 2.3 of the CPE specification
        target_sw: Characterizes the software computing environment within which
            the product operates. Extended attribute introduced with
            version 2.3 of the CPE specification
        hardware_sw: Characterizes the instruction set architecture (e.g., x86)
            on which the product operates. Extended attribute introduced with
            version 2.3 of the CPE specification
        other: Captures any other general descriptive or identifying information
            which is vendor- or product-specific and which does not logically
            fit in any other attribute value. Extended attribute introduced with
            version 2.3 of the CPE specification
        cpe_string: The original parsed CPE string

    Example:
        .. code-block:: python

            from pontos.cpe import CPE

            cpe = CPE.from_string("cpe:2.3:o:google:android:13.0:*:*:*:*:*:*:*")

            print(cpe.vendor)           # google
            print(cpe.product)          # android
            print(cpe.version)          # 13.0
            print(cpe.as_uri_binding()) # cpe:/o:google:android:13.0
    """

    def __init__(
        self,
        *,
        cpe_string: Optional[str] = None,
        part: Part,
        vendor: str,
        product: str,
        version: Optional[str] = None,
        update: Optional[str] = None,
        edition: Optional[str] = None,
        language: Optional[str] = None,
        sw_edition: Optional[str] = None,
        target_sw: Optional[str] = None,
        target_hw: Optional[str] = None,
        other: Optional[str] = None,
    ) -> None:
        self.cpe_string = cpe_string
        self.__wnf__ = CPEWellFormed(
            part=part,
            vendor=vendor,
            product=product,
            version=version,
            update=update,
            edition=edition,
            language=language,
            sw_edition=sw_edition,
            target_sw=target_sw,
            target_hw=target_hw,
            other=other,
        )
        self.part = part
        self.vendor = unquote_attribute_value(vendor)
        self.product = unquote_attribute_value(product)
        self.version = unquote_attribute_value(version)
        self.update = unquote_attribute_value(update)
        self.edition = unquote_attribute_value(edition)
        self.language = unquote_attribute_value(language)
        self.sw_edition = unquote_attribute_value(sw_edition)
        self.target_sw = unquote_attribute_value(target_sw)
        self.target_hw = unquote_attribute_value(target_hw)
        self.other = unquote_attribute_value(other)

    @staticmethod
    def from_string(cpe: str) -> "CPE":
        """
        Create a new CPE from a string
        """
        cleaned_cpe = cpe.strip().lower()
        parts = split_cpe(cleaned_cpe)

        if is_uri_binding(cleaned_cpe):
            values: dict[str, Optional[str]] = dict(
                zip(
                    [
                        "vendor",
                        "product",
                        "version",
                        "update",
                        "edition",
                        "language",
                    ],
                    parts[2:],
                )
            )
            for attribute in [
                "vendor",
                "product",
                "version",
                "update",
                "language",
            ]:
                values[attribute] = unbind_value_uri(values.get(attribute))

            edition = values.get("edition")
            if (
                edition is None
                or edition == ""
                or edition == NA
                or edition[0] != "~"
            ):
                edition = unbind_value_uri(edition)
            else:
                values.update(unpack_edition(edition))

            return CPE(cpe_string=cleaned_cpe, part=Part(parts[1][1]), **values)  # type: ignore[arg-type]

        elif is_formatted_string_binding(cleaned_cpe):
            values = dict(
                zip(
                    [
                        "vendor",
                        "product",
                        "version",
                        "update",
                        "edition",
                        "language",
                        "sw_edition",
                        "target_sw",
                        "target_hw",
                        "other",
                    ],
                    [unbind_value_from_formatted_string(a) for a in parts[3:]],
                )
            )

            return CPE(cpe_string=cleaned_cpe, part=Part(parts[2]), **values)  # type: ignore[arg-type]

        raise CPEParsingError(
            f"Invalid CPE string '{cpe}'. CPE does not start with "
            "'cpe:/' or 'cpe:2.3'"
        )

    def has_extended_attribute(self) -> bool:
        """
        Returns True if the CPE has an extended attribute set
        """
        return bool(
            self.sw_edition or self.target_sw or self.target_hw or self.other
        )

    def is_uri_binding(self) -> bool:
        """
        Returns True if the CPE is parsed from a URI binding
        """
        if self.cpe_string:
            return is_uri_binding(self.cpe_string)
        return not self.has_extended_attribute()

    def is_formatted_string_binding(self) -> bool:
        """
        Returns True if the CPE is parsed from a formatted string binding
        """
        if self.cpe_string:
            return is_formatted_string_binding(self.cpe_string)
        return self.has_extended_attribute()

    def as_uri_binding(self) -> str:
        """
        Converts the CPE to an URI binding
        """
        part = self.part.value
        vendor = bind_value_for_uri(self.__wnf__.vendor)
        product = bind_value_for_uri(self.__wnf__.product)
        version = bind_value_for_uri(self.__wnf__.version)
        update = bind_value_for_uri(self.__wnf__.update)
        language = bind_value_for_uri(self.__wnf__.language)
        edition = bind_value_for_uri(self.__wnf__.edition)
        sw_edition = bind_value_for_uri(self.__wnf__.sw_edition)
        target_sw = bind_value_for_uri(self.__wnf__.target_sw)
        target_hw = bind_value_for_uri(self.__wnf__.target_hw)
        other = bind_value_for_uri(self.__wnf__.other)

        edition = pack_extended_attributes(
            edition,
            sw_edition,
            target_sw,
            target_hw,
            other,
        )

        uri = f"cpe:/{part}:{vendor}:{product}"
        if version or update or edition or language:
            uri = f"{uri}:{version}"
        if update or edition or language:
            uri = f"{uri}:{update}"
        if edition or language:
            uri = f"{uri}:{edition}"
        if language:
            uri = f"{uri}:{language}"
        return uri

    def as_formatted_string_binding(self) -> str:
        """
        Converts the CPE to a formatted string binding
        """
        part = self.part.value
        vendor = bind_value_for_formatted_string(self.__wnf__.vendor)
        product = bind_value_for_formatted_string(self.__wnf__.product)
        version = bind_value_for_formatted_string(self.__wnf__.version)
        update = bind_value_for_formatted_string(self.__wnf__.update)
        edition = bind_value_for_formatted_string(self.__wnf__.edition)
        language = bind_value_for_formatted_string(self.__wnf__.language)
        sw_edition = bind_value_for_formatted_string(self.__wnf__.sw_edition)
        target_sw = bind_value_for_formatted_string(self.__wnf__.target_sw)
        target_hw = bind_value_for_formatted_string(self.__wnf__.target_hw)
        other = bind_value_for_formatted_string(self.__wnf__.other)
        return (
            f"cpe:2.3:{part}:{vendor}:{product}:{version}:{update}:"
            f"{edition}:{language}:{sw_edition}:{target_sw}:{target_hw}:{other}"
        )

    def clone(
        self,
        **kwargs,
    ) -> "CPE":
        """
        Clone a CPE and allow to override parts

        Example:
            .. code-block:: python

                from pontos.cpe import CPE, ANY

                android_13 = CPE.from_string(
                    "cpe:2.3:o:google:android:13.0:*:*:*:*:*:*:*"
                )
                all_android_versions = cpe.clone(version=ANY)
        """
        args = {
            "part": self.__wnf__.part,
            "vendor": self.__wnf__.vendor,
            "product": self.__wnf__.product,
            "version": self.__wnf__.version,
            "update": self.__wnf__.update,
            "edition": self.__wnf__.edition,
            "language": self.__wnf__.language,
            "sw_edition": self.__wnf__.sw_edition,
            "target_sw": self.__wnf__.target_sw,
            "target_hw": self.__wnf__.target_hw,
            "other": self.__wnf__.other,
            "cpe_string": self.cpe_string,
        }
        args.update(**kwargs)
        return CPE(**args)  # type: ignore[arg-type]

    def __str__(self) -> str:
        """
        Returns the string representation (uri of formatted string) of the CPE
        """
        if self.cpe_string:
            return self.cpe_string

        if not self.has_extended_attribute():
            return self.as_uri_binding()

        return self.as_formatted_string_binding()

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f'part="{self.part}" '
            f'vendor="{self.vendor}" '
            f'product="{self.product}" '
            f'version="{self.version}" '
            f'update="{self.update}" '
            f'edition="{self.edition}" '
            f'language="{self.language}" '
            f'sw_edition="{self.sw_edition}" '
            f'target_sw="{self.target_sw}" '
            f'target_hw="{self.target_hw}" '
            f'other="{self.other}"'
            ">"
        )

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, CPE):
            return False
        return str(self) == str(other)



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from dataclasses import dataclass, field

from pontos.github.models.base import GitHubModel, User
from pontos.github.models.organization import Repository
from pontos.models import StrEnum

__all__ = [
    "PackageType",
    "PackageVisibility",
    "Package",
    "Container",
    "PackageVersionMetadata",
    "PackageVersion",
]


class PackageType(StrEnum):
    CONTAINER = "container"
    DOCKER = "docker"
    MAVEN = "maven"
    NPM = "npm"
    NUGET = "nuget"
    RUBYGEMS = "rubygems"


class PackageVisibility(StrEnum):
    PUBLIC = "public"
    PRIVATE = "private"


@dataclass
class Package(GitHubModel):
    id: int
    name: str
    package_type: PackageType
    owner: User
    version_count: int
    visibility: PackageVisibility
    url: str
    created_at: str
    updated_at: str
    repository: Repository
    html_url: str


@dataclass
class Container(GitHubModel):
    tags: list[str] = field(default_factory=list)


@dataclass
class PackageVersionMetadata(GitHubModel):
    package_type: PackageType
    container: Container


@dataclass
class PackageVersion(GitHubModel):
    id: int
    name: str
    url: str
    package_html_url: str
    created_at: str
    updated_at: str
    html_url: str
    metadata: PackageVersionMetadata



# SPDX-FileCopyrightText: 2021-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import os
import re
import sys
import warnings
from contextlib import asynccontextmanager, contextmanager
from datetime import timedelta
from enum import Enum
from functools import wraps
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    AsyncContextManager,
    AsyncIterator,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterator,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import httpx

from pontos.errors import PontosError
from pontos.typing import SupportsStr

__all__ = (
    "AsyncDownloadProgressIterable",
    "DownloadProgressIterable",
    "add_sys_path",
    "deprecated",
    "download_async",
    "download",
    "ensure_unload_module",
    "enum_or_value",
    "parse_timedelta",
    "snake_case",
    "unload_module",
)

DEFAULT_TIMEOUT = 1000
DEFAULT_CHUNK_SIZE = 4096


async def upload(file_path: Path) -> AsyncIterator[bytes]:
    with file_path.open("rb") as f:
        read = f.read(DEFAULT_CHUNK_SIZE)
        while read:
            yield read
            read = f.read(DEFAULT_CHUNK_SIZE)


T = TypeVar("T", str, bytes)


class AsyncDownloadProgressIterable(Generic[T]):
    """
    An async iterator to iterate over a downloadable content and the progress.

    Example:
        .. code-block:: python

            from pontos.helper import AsyncDownloadProgressIterable

            it = AsyncDownloadProgressIterable(...)
            async for content, progress in it:
                file.write(content)
                print(progress)
    """

    def __init__(
        self,
        *,
        content_iterator: AsyncIterator[T],
        url: SupportsStr,
        length: Optional[int],
    ):
        """
        Create a new AsyncDownloadProgressIterable instance

        Args:
            content_iterator: An async iterator to call for getting the content.
                Should be a stream of bytes or strings.
            url: The URL where the content gets downloaded.
            length: Length of the content.
        """
        self._content_iterator: AsyncIterator[T] = content_iterator
        self._url = str(url)
        self._length = None if length is None else int(length)

    @property
    def length(self) -> Optional[int]:
        """
        Size in bytes of the to be downloaded file or None if the size is not
        available
        """
        return self._length

    @property
    def url(self) -> str:
        """
        The URL where the content gets downloaded from
        """
        return self._url

    async def _download(self) -> AsyncIterator[Tuple[T, Optional[float]]]:
        dl = 0
        async for content in self._content_iterator:
            dl += len(content)
            progress = dl / self._length * 100 if self._length else None
            yield content, progress

    def __aiter__(self) -> AsyncIterator[Tuple[T, Optional[float]]]:
        """
        Returns an async iterator yielding a tuple of content and progress.
        The progress is expressed as percent of the content length.
        """
        return self._download()


@asynccontextmanager
async def download_async(
    stream: AsyncContextManager[httpx.Response],
    *,
    content_length: Optional[int] = None,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    url: Optional[str] = None,
) -> AsyncIterator[AsyncDownloadProgressIterable[bytes]]:
    """
    An async context manager that returns an AsyncDownloadProgressIterable.

    It ensures that the stream is closed automatically via the context manager.

    Args:
        stream: An async context manager providing a streaming response.
        content_length: Optional length of the content to download. If now
            provided it is determined from the response if available.
        chunk_size: Download the content in chunks of this size.
        url: Use a specific URL. If not set the URL of the response is used.

    Returns:
        A context manager containing an AsyncDownloadProgressIterable

    Raises:
        HTTPStatusError: If the request was invalid

    Example:
        .. code-block:: python

            import httpx
            from pontos.helper import download_async

            client = httpx.AsyncClient(...)
            stream = client.stream("GET, "https://foo.bar/baz.zip)

            async with download_async(stream) as download:
                async for content, progress in download:
                    file.write(content)
                    print(progress)
    """
    async with stream as response:
        response.raise_for_status()

        if not content_length:
            content_length = response.headers.get("content-length")

        yield AsyncDownloadProgressIterable(
            url=url if url else response.url,
            content_iterator=response.aiter_bytes(chunk_size=chunk_size),
            length=content_length,
        )


class DownloadProgressIterable:
    """
    An synchronous iterator to iterate over a download progress.

    Example:
        .. code-block:: python

            from pontos.helper import DownloadProgressIterable

            it = DownloadProgressIterable(...)
            for progress in it:
                print(progress)
    """

    def __init__(
        self,
        *,
        content_iterator: Iterator[bytes],
        url: str,
        destination: Path,
        length: Optional[int],
    ) -> None:
        """
        Create a new DownloadProgressIterable instance

        Args:
            content_iterator: An iterator of bytes to write to destination path
            url: A URL where the content will be downloaded from
            destination: Path to write the downloaded content to
            length: Length of the content to be downloaded
        """
        self._content_iterator = content_iterator
        self._url = url
        self._destination = destination
        self._length = None if length is None else int(length)

    @property
    def length(self) -> Optional[int]:
        """
        Size in bytes of the to be downloaded file or None if the size is not
        available
        """
        return self._length

    @property
    def destination(self) -> Path:
        """
        Destination path of the to be downloaded file
        """
        return self._destination

    @property
    def url(self) -> str:
        return self._url

    def _download(self) -> Iterator[Optional[float]]:
        dl = 0
        with self._destination.open("wb") as f:
            for content in self._content_iterator:
                dl += len(content)
                f.write(content)
                yield dl / self._length if self._length else None

    def __iter__(self) -> Iterator[Optional[float]]:
        return self._download()

    def run(self):
        """
        Just run the download without caring about the progress
        """
        try:
            it = iter(self)
            while True:
                next(it)
        except StopIteration:
            pass


@contextmanager
def download(
    url: str,
    destination: Optional[Union[Path, str]] = None,
    *,
    headers: Optional[Dict[str, Any]] = None,
    params: Optional[Dict[str, Any]] = None,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    timeout: int = DEFAULT_TIMEOUT,
) -> Generator[DownloadProgressIterable, None, None]:
    """Download file in url to filename

    Arguments:
        url: The url of the file we want to download
        destination: Path of the file to store the download in. If set it will
                     be derived from the passed URL.
        headers: HTTP headers to use for the download
        params: HTTP request parameters to use for the download
        chunk_size: Download file in chunks of this size
        timeout: Connection timeout

    Raises:
        HTTPStatusError: If the request was invalid

    Returns:
        A DownloadProgressIterator that yields the progress of the download in
        percent for each downloaded chunk or None for each chunk if the progress
        is unknown.

    Example:
        .. code-block:: python

            from pontos.helper import download

            with download("https://example.com/some/file") as progress_it:
                for progress in progress_it:
                    print(progress)
    """
    destination = (
        Path(url.split("/")[-1]) if not destination else Path(destination)
    )

    with httpx.stream(
        "GET",
        url,
        timeout=timeout,
        follow_redirects=True,
        headers=headers,
        params=params,
    ) as response:
        response.raise_for_status()

        total_length = response.headers.get("content-length")

        yield DownloadProgressIterable(
            url=url,
            content_iterator=response.iter_bytes(chunk_size=chunk_size),
            destination=destination,
            length=total_length,
        )


def deprecated(
    _func_or_cls: Union[str, Callable, Type, None] = None,
    *,
    since: Optional[str] = None,
    reason: Optional[str] = None,
):
    """
    A decorator to mark functions, classes and methods as deprecated

    Args:
        since: An optional version since the referenced item is deprecated.
        reason: An optional reason why the references item is deprecated.

    Examples:
        .. code-block:: python

            from pontos.helper import deprecated

            @deprecated
            def my_function(*args, **kwargs):
                ...

            @deprecated("The function is obsolete. Please use my_func instead.")
            def my_function(*args, **kwargs):
                ...

            @deprecated(
                since="1.2.3",
                reason="The function is obsolete. Please use my_func instead."
            )
            def my_function(*args, **kwargs):
                ...

            @deprecated(reason="The class will be removed in version 3.4.5")
            class Foo:
                ...

            class Foo:
                @deprecated(since="2.3.4")
                def bar(self, *args, **kwargs):
                    ...
    """
    if isinstance(_func_or_cls, str):
        reason = _func_or_cls
        _func_or_cls = None

    def decorator_repeat(func_or_cls):
        module = func_or_cls.__module__
        name = func_or_cls.__name__

        if module == "__main__":
            msg = f"{name} is deprecated."
        else:
            msg = f"{module}.{name} is deprecated."

        if since:
            msg += f" It is deprecated since version {since}."
        if reason:
            msg += f" {reason}"

        @wraps(func_or_cls)
        def wrapper(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning, stacklevel=3)
            return func_or_cls(*args, **kwargs)

        return wrapper

    if _func_or_cls is None:
        return decorator_repeat
    else:
        return decorator_repeat(_func_or_cls)


@contextmanager
def add_sys_path(
    directory: Union[str, os.PathLike]
) -> Generator[None, None, None]:
    """
    Context Manager to add a directory path to the module search path aka.
    sys.path. The directory path is removed when the context manager is left.

    Args:
        directory: A os.PathLike directory to add to sys.path

    Example:
        .. code-block:: python

            from pontos.helper import add_sys_path

            with add_sys_path("/tmp/test-modules"):
                import mymodule
    """
    directory = os.fspath(directory)

    if sys.path[0] != directory:
        sys.path.insert(0, directory)

    try:
        yield
    finally:
        try:
            sys.path.remove(directory)
        except ValueError:
            # directory was not in the path
            pass


def unload_module(module: Union[str, ModuleType]) -> None:
    """
    Unload a Python module

    Args:
        module: Module instance or name of the Python module to unload.
            For example: foo.bar
    """
    name = module.__name__ if isinstance(module, ModuleType) else module

    if name in sys.modules:
        del sys.modules[name]


@contextmanager
def ensure_unload_module(
    module: Union[str, ModuleType]
) -> Generator[None, None, None]:
    """
    A context manager to ensure that a module gets removed even if an error
    occurs

    Args:
        module: Module instance or name of the Python module to unload.
            For example: foo.bar

    Example:
        .. code-block:: python

            from pontos.helper import ensure_unload_module

            with ensure_unload_module("foo.bar"):
                do_something()
    """
    try:
        yield
    finally:
        unload_module(module)


def snake_case(value: str) -> str:
    """
    Convert a string to snake case/underscore naming scheme

    Args:
        value: String to convert into snake case

    Example:
        .. code-block:: python

            from pontos.helper import snake_case

            snake_case("CamelCase")

        will return "camel_case"
    """
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", value)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def enum_or_value(value: Union[Enum, Any]) -> Any:
    """
    Return the value of an Enum or the value if it isn't an Enum
    """
    if isinstance(value, Enum):
        return value.value

    return value


regex = re.compile(
    r"^((?P<weeks>[\.\d]+?)w)?((?P<days>[\.\d]+?)d)?((?P<hours>[\.\d]+?)h)?"
    r"((?P<minutes>[\.\d]+?)m)?((?P<seconds>[\.\d]+?)s)?$"
)


def parse_timedelta(time_str: str) -> timedelta:
    """
    Parse a timedelta from a string

    Examples:
        .. code-block:: python

            from pontos.helper import parse_timedelta

            parse_timedelta("1.5h")
            parse_timedelta("1w2d4h5m6s")
    """
    time_match = regex.match(time_str)
    if not time_match:
        raise PontosError(f"Invalid timedelta format '{time_str}'.")

    parts = time_match.groupdict()
    time_params = {}
    for name, param in parts.items():
        if param:
            time_params[name] = float(param)

    return timedelta(**time_params)



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from collections import defaultdict
from pathlib import Path
from typing import AsyncIterator, Dict, Iterable, List, Optional, Union

from pontos.github.api.client import GitHubAsyncREST
from pontos.github.api.helper import JSON_OBJECT
from pontos.github.models.base import FileStatus
from pontos.github.models.pull_request import (
    Comment,
    PullRequest,
    PullRequestCommit,
)


class GitHubAsyncRESTPullRequests(GitHubAsyncREST):
    async def exists(self, repo: str, pull_request: Union[int, str]) -> bool:
        """
        Check if a single branch in a repository exists

        Args:
            repo: GitHub repository (owner/name) to use
            pull_request: Pull request number to check

        Returns:
            True if the pull requests exists

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    exists = await api.pull_request.exists("foo/bar", 123)
        """
        api = f"/repos/{repo}/pulls/{pull_request}"
        response = await self._client.get(api)
        return response.is_success

    async def get(
        self, repo: str, pull_request: Union[int, str]
    ) -> PullRequest:
        """
        Get information about a pull request

        https://docs.github.com/en/rest/pulls/pulls#get-a-pull-request

        Args:
            repo: GitHub repository (owner/name) to use
            pull_request: Pull request number

        Returns:
            Information about the pull request

        Raises:
            httpx.HTTPStatusError: If the request was invalid

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    pr = await api.pull_requests.get("foo/bar", 123)
                    print(pr)
        """
        api = f"/repos/{repo}/pulls/{pull_request}"
        response = await self._client.get(api)
        response.raise_for_status()
        return PullRequest.from_dict(response.json())

    async def commits(
        self, repo: str, pull_request: Union[int, str]
    ) -> AsyncIterator[PullRequestCommit]:
        """
        Get all commit information of a pull request

        https://docs.github.com/en/rest/pulls/pulls#list-commits-on-a-pull-request

        Hint: At maximum GitHub allows to receive 250 commits of a pull request.

        Args:
            repo: GitHub repository (owner/name) to use
            pull_request: Pull request number

        Returns:
            An async iterator yielding pull request commits

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    async for commit in api.pull_requests.commits(
                        "foo/bar", 123
                    ):
                        print(commit)
        """
        # per default github only shows 35 commits and at max it is only
        # possible to receive 100
        params = {"per_page": "100"}
        api = f"/repos/{repo}/pulls/{pull_request}/commits"

        async for response in self._client.get_all(api, params=params):
            for commit in response.json():
                yield PullRequestCommit.from_dict(commit)

    async def create(
        self,
        repo: str,
        *,
        head_branch: str,
        base_branch: str,
        title: str,
        body: str,
    ) -> PullRequest:
        """
        Create a new Pull Request on GitHub

        https://docs.github.com/en/rest/pulls/pulls#create-a-pull-request

        Args:
            repo: GitHub repository (owner/name) to use
            head_branch: Branch to create a pull request from
            base_branch: Branch as target for the pull request
            title: Title for the pull request
            body: Description for the pull request. Can be formatted in Markdown

        Raises:
            httpx.HTTPStatusError if the request was invalid

        Returns:
            A new pull request

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    pr = await api.pull_requests.create(
                        "foo/bar",
                        head_branch="a-new-feature",
                        base_branch="main",
                        title="A new Feature is ready",
                        body="Created a new feature",
                    )
        """
        api = f"/repos/{repo}/pulls"
        data: JSON_OBJECT = {
            "head": head_branch,
            "base": base_branch,
            "title": title,
            "body": body.replace("\\n", "\n"),
        }
        response = await self._client.post(api, data=data)
        response.raise_for_status()
        return PullRequest.from_dict(response.json())

    async def update(
        self,
        repo: str,
        pull_request: Union[int, str],
        *,
        base_branch: Optional[str] = None,
        title: Optional[str] = None,
        body: Optional[str] = None,
    ) -> PullRequest:
        """
        Update a Pull Request on GitHub

        https://docs.github.com/en/rest/pulls/pulls#update-a-pull-request

        Args:
            repo: GitHub repository (owner/name) to use
            pull_request: Pull request number
            base_branch: Branch as target for the pull request. Leave empty for
                keeping the current one.
            title: Title for the pull request. Leave empty for keeping the
                current one.
            body: Description for the pull request. Can be formatted in
                Markdown. Leave empty for keeping the current one.
        Raises:
            httpx.HTTPStatusError if the request was invalid

        Returns:
            Updated pull request

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    pr = await api.pull_requests.update(
                        "foo/bar",
                        123,
                        title="Another new Feature",
                    )
        """
        api = f"/repos/{repo}/pulls/{pull_request}"

        data: JSON_OBJECT = {}
        if base_branch:
            data["base"] = base_branch
        if title:
            data["title"] = title
        if body:
            data["body"] = body.replace("\\n", "\n")

        response = await self._client.post(api, data=data)
        response.raise_for_status()
        return PullRequest.from_dict(response.json())

    async def add_comment(
        self, repo: str, pull_request: Union[int, str], comment: str
    ) -> Comment:
        """
        Add a comment to a pull request on GitHub

        https://docs.github.com/en/rest/issues/comments#create-an-issue-comment

        Args:
            repo: GitHub repository (owner/name) to use
            pull_request: Pull request number where to add a comment
            comment: The actual comment message. Can be formatted in Markdown.

        Raises:
            httpx.HTTPStatusError if the request was invalid

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    await api.pull_requests.add_comment(
                        "foo/bar",
                        123,
                        "A new comment for the pull request",
                    )
        """
        api = f"/repos/{repo}/issues/{pull_request}/comments"
        data: JSON_OBJECT = {"body": comment}
        response = await self._client.post(api, data=data)
        response.raise_for_status()
        return Comment.from_dict(response.json())

    async def update_comment(
        self, repo: str, comment_id: Union[str, int], comment: str
    ) -> Comment:
        """
        Update a comment to a pull request on GitHub

        https://docs.github.com/en/rest/issues/comments#update-an-issue-comment

        Args:
            repo: GitHub repository (owner/name) to use
            comment_id: The unique identifier of the comment
            comment: The actual comment message. Can be formatted in Markdown.

        Raises:
            httpx.HTTPStatusError if the request was invalid

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    await api.pull_requests.update_comment(
                        "foo/bar",
                        123,
                        "A new comment for the pull request",
                    )
        """
        api = f"/repos/{repo}/issues/comments/{comment_id}"
        data: JSON_OBJECT = {"body": comment}
        response = await self._client.post(api, data=data)
        response.raise_for_status()
        return Comment.from_dict(response.json())

    async def comments(
        self, repo: str, pull_request: Union[int, str]
    ) -> AsyncIterator[Comment]:
        """
        Get all comments of a pull request on GitHub

        https://docs.github.com/en/rest/issues/comments#list-issue-comments

        Args:
            repo: GitHub repository (owner/name) to use
            pull_request: Pull request number where to add a comment

        Raises:
            httpx.HTTPStatusError if the request was invalid

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    async for comment in api.pull_requests.comments(
                        "foo/bar",
                        123,
                    ):
                        print(comment)
        """
        params = {"per_page": "100"}
        api = f"/repos/{repo}/issues/{pull_request}/comments"

        async for response in self._client.get_all(api, params=params):
            response.raise_for_status()

            for comment in response.json():
                yield Comment.from_dict(comment)

    async def files(
        self,
        repo: str,
        pull_request: Union[int, str],
        *,
        status_list: Optional[Iterable[FileStatus]] = None,
    ) -> Dict[FileStatus, Iterable[Path]]:
        """
        Get files of a pull request

        https://docs.github.com/en/rest/pulls/pulls#list-pull-requests-files

        Hint: At maximum GitHub allows to receive 3000 files of a commit.

        Args:
            repo: GitHub repository (owner/name) to use
            pull_request: Pull request number
            status_list: Optional iterable of status change types that should be
                included in the response

        Returns:
            Information about the files in the pull request as a dict

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi
                from pontos.github.models import FileStatus

                async with GitHubAsyncRESTApi(token) as api:
                    status = await api.pull_requests.files("foo/bar", 123)
                    # list changed files
                    print(status[FileStatus.MODIFIED])
        """
        # per default github only shows 35 files per page and at max it is only
        # possible to receive 100
        params = {"per_page": "100"}
        api = f"/repos/{repo}/pulls/{pull_request}/files"
        file_dict: Dict[FileStatus, List[Path]] = defaultdict(list)

        async for response in self._client.get_all(api, params=params):
            response.raise_for_status()

            for f in response.json():
                try:
                    status = FileStatus(f["status"])
                except ValueError:
                    # unknown status
                    continue

                if not status_list or status in status_list:
                    file_dict[status].append(Path(f["filename"]))

        return file_dict  # type: ignore



# SPDX-FileCopyrightText: 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass
from typing import Optional

from pontos.github.models.base import GitHubModel


@dataclass
class ActionsMinutesUsedBreakdown(GitHubModel):
    """
    Attributes:
        UBUNTU: Total minutes used on Ubuntu runner machines
        MACOS: Total minutes used on macOS runner machines
        WINDOWS: Total minutes used on Windows runner machines
        total: Total minutes used on all runner machines
    """

    UBUNTU: Optional[int] = None
    MACOS: Optional[int] = None
    WINDOWS: Optional[int] = None
    total: Optional[int] = None


@dataclass
class ActionsBilling(GitHubModel):
    """
    Billing Information for using GitHub Actions

    Attributes:
        total_minutes_used: The sum of the free and paid GitHub Actions minutes
            used
        total_paid_minutes_used: The total paid GitHub Actions minutes used
        included_minutes: The amount of free GitHub Actions minutes available
        minutes_used_breakdown:
    """

    total_minutes_used: int
    total_paid_minutes_used: int
    included_minutes: int
    minutes_used_breakdown: ActionsMinutesUsedBreakdown


@dataclass
class PackagesBilling(GitHubModel):
    """
    Billing Information for using GitHub Packages

    Attributes:
        total_gigabytes_bandwidth_used: Sum of the free and paid storage space
            (GB) for GitHub Packages
        total_paid_gigabytes_bandwidth_used: Total paid storage space (GB) for
            GitHub Packages
        included_gigabytes_bandwidth: Free storage space (GB) for GitHub
            Packages
    """

    total_gigabytes_bandwidth_used: int
    total_paid_gigabytes_bandwidth_used: int
    included_gigabytes_bandwidth: int


@dataclass
class StorageBilling(GitHubModel):
    """
    Billing Information for using GitHub storage

    Attributes:
        days_left_in_billing_cycle: Numbers of days left in billing cycle
        estimated_paid_storage_for_month: Estimated storage space (GB) used in
            billing cycle
        estimated_storage_for_month: Estimated sum of free and paid storage
            space (GB) used in billing cycle
    """

    days_left_in_billing_cycle: int
    estimated_paid_storage_for_month: int
    estimated_storage_for_month: int



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# pylint: disable=line-too-long

import unittest
from datetime import datetime, timezone

from pontos.github.models.artifact import Artifact


class ArtifactTestCase(unittest.TestCase):
    def test_from_dict(self):
        artifact = Artifact.from_dict(
            {
                "id": 1,
                "node_id": "MDg6QXJ0aWZhY3QxMQ==",
                "name": "Rails",
                "size_in_bytes": 556,
                "url": "https://api.github.com/repos/octo-org/octo-docs/actions/artifacts/11",
                "archive_download_url": "https://api.github.com/repos/octo-org/octo-docs/actions/artifacts/11/zip",
                "expired": False,
                "created_at": "2020-01-10T14:59:22Z",
                "expires_at": "2020-03-21T14:59:22Z",
                "updated_at": "2020-02-21T14:59:22Z",
                "workflow_run": {
                    "id": 1,
                    "repository_id": 2,
                    "head_repository_id": 3,
                    "head_branch": "main",
                    "head_sha": "328faa0536e6fef19753d9d91dc96a9931694ce3",
                },
            }
        )

        self.assertEqual(artifact.id, 1)
        self.assertEqual(artifact.node_id, "MDg6QXJ0aWZhY3QxMQ==")
        self.assertEqual(artifact.name, "Rails")
        self.assertEqual(artifact.size_in_bytes, 556)
        self.assertEqual(
            artifact.url,
            "https://api.github.com/repos/octo-org/octo-docs/actions/artifacts/11",
        )
        self.assertEqual(
            artifact.archive_download_url,
            "https://api.github.com/repos/octo-org/octo-docs/actions/artifacts/11/zip",
        )
        self.assertFalse(artifact.expired)
        self.assertEqual(
            artifact.created_at,
            datetime(2020, 1, 10, 14, 59, 22, tzinfo=timezone.utc),
        )
        self.assertEqual(
            artifact.expires_at,
            datetime(2020, 3, 21, 14, 59, 22, tzinfo=timezone.utc),
        )
        self.assertEqual(
            artifact.updated_at,
            datetime(2020, 2, 21, 14, 59, 22, tzinfo=timezone.utc),
        )

        workflow_run = artifact.workflow_run
        self.assertEqual(workflow_run.id, 1)
        self.assertEqual(workflow_run.repository_id, 2)
        self.assertEqual(workflow_run.head_repository_id, 3)
        self.assertEqual(workflow_run.head_branch, "main")
        self.assertEqual(
            workflow_run.head_sha, "328faa0536e6fef19753d9d91dc96a9931694ce3"
        )



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# pylint: disable=redefined-builtin, line-too-long

from unittest.mock import MagicMock

import httpx

from pontos.github.api.teams import GitHubAsyncRESTTeams, TeamPrivacy, TeamRole
from pontos.github.models.base import Permission
from tests import AsyncIteratorMock, aiter, anext
from tests.github.api import GitHubAsyncRESTTestCase, create_response


class GitHubAsyncRESTTeamsTestCase(GitHubAsyncRESTTestCase):
    api_cls = GitHubAsyncRESTTeams

    async def test_get_all(self):
        response1 = create_response()
        response1.json.return_value = [
            {
                "id": 1,
                "node_id": "MDQ6VGVhbTE=",
                "url": "https://api.github.com/teams/1",
                "html_url": "https://github.com/orgs/github/teams/justice-league",
                "name": "Justice League",
                "slug": "justice-league",
                "description": "A great team.",
                "privacy": "closed",
                "permission": "admin",
                "members_url": "https://api.github.com/teams/1/members{/member}",
                "repositories_url": "https://api.github.com/teams/1/repos",
                "parent": None,
            }
        ]
        response2 = create_response()
        response2.json.return_value = [
            {
                "id": 2,
                "node_id": "MDQ6VGVhbTE=",
                "url": "https://api.github.com/teams/1",
                "html_url": "https://github.com/orgs/github/teams/justice-league",
                "name": "Justice League",
                "slug": "justice-league",
                "description": "A great team.",
                "privacy": "closed",
                "permission": "admin",
                "members_url": "https://api.github.com/teams/1/members{/member}",
                "repositories_url": "https://api.github.com/teams/1/repos",
                "parent": None,
            },
            {
                "id": 3,
                "node_id": "MDQ6VGVhbTE=",
                "url": "https://api.github.com/teams/1",
                "html_url": "https://github.com/orgs/github/teams/justice-league",
                "name": "Justice League",
                "slug": "justice-league",
                "description": "A great team.",
                "privacy": "closed",
                "permission": "admin",
                "members_url": "https://api.github.com/teams/1/members{/member}",
                "repositories_url": "https://api.github.com/teams/1/repos",
                "parent": None,
            },
        ]

        self.client.get_all.return_value = AsyncIteratorMock(
            [response1, response2]
        )

        async_it = aiter(self.api.get_all("foo"))
        team = await anext(async_it)
        self.assertEqual(team.id, 1)
        team = await anext(async_it)
        self.assertEqual(team.id, 2)
        team = await anext(async_it)
        self.assertEqual(team.id, 3)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/orgs/foo/teams",
            params={"per_page": "100"},
        )

    async def test_create(self):
        response = create_response()
        response.json.return_value = {
            "id": 1,
            "node_id": "MDQ6VGVhbTE=",
            "url": "https://api.github.com/teams/1",
            "html_url": "https://github.com/orgs/github/teams/justice-league",
            "name": "Justice League",
            "slug": "justice-league",
            "description": "A great team.",
            "privacy": "closed",
            "permission": "admin",
            "members_url": "https://api.github.com/teams/1/members{/member}",
            "repositories_url": "https://api.github.com/teams/1/repos",
            "parent": None,
        }
        self.client.post.return_value = response

        team = await self.api.create(
            "foo",
            "bar",
            description="A description",
            maintainers=["foo", "bar"],
            repo_names=["foo/bar", "foo/baz"],
            privacy=TeamPrivacy.CLOSED,
            parent_team_id="123",
        )

        self.client.post.assert_awaited_once_with(
            "/orgs/foo/teams",
            data={
                "name": "bar",
                "description": "A description",
                "maintainers": ["foo", "bar"],
                "repo_names": ["foo/bar", "foo/baz"],
                "privacy": "closed",
                "parent_team_id": "123",
            },
        )

        self.assertEqual(team.id, 1)

    async def test_create_failure(self):
        response = create_response()
        self.client.post.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.create(
                "foo",
                "bar",
                description="A description",
                maintainers=["foo", "bar"],
                repo_names=["foo/bar", "foo/baz"],
                privacy=TeamPrivacy.CLOSED,
                parent_team_id="123",
            )

        self.client.post.assert_awaited_once_with(
            "/orgs/foo/teams",
            data={
                "name": "bar",
                "description": "A description",
                "maintainers": ["foo", "bar"],
                "repo_names": ["foo/bar", "foo/baz"],
                "privacy": "closed",
                "parent_team_id": "123",
            },
        )

    async def test_get(self):
        response = create_response()
        response.json.return_value = {
            "id": 1,
            "node_id": "MDQ6VGVhbTE=",
            "url": "https://api.github.com/teams/1",
            "html_url": "https://github.com/orgs/github/teams/justice-league",
            "name": "Justice League",
            "slug": "justice-league",
            "description": "A great team.",
            "privacy": "closed",
            "permission": "admin",
            "members_url": "https://api.github.com/teams/1/members{/member}",
            "repositories_url": "https://api.github.com/teams/1/repos",
            "parent": None,
        }
        self.client.get.return_value = response

        team = await self.api.get("foo", "bar")

        self.client.get.assert_awaited_once_with(
            "/orgs/foo/teams/bar",
        )

        self.assertEqual(team.id, 1)

    async def test_get_failure(self):
        response = create_response()
        self.client.get.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.get("foo", "bar")

        self.client.get.assert_awaited_once_with(
            "/orgs/foo/teams/bar",
        )

    async def test_update(self):
        response = create_response()
        response.json.return_value = {
            "id": 1,
            "node_id": "MDQ6VGVhbTE=",
            "url": "https://api.github.com/teams/1",
            "html_url": "https://github.com/orgs/github/teams/justice-league",
            "name": "baz",
            "slug": "justice-league",
            "description": "A description",
            "privacy": "closed",
            "permission": "admin",
            "members_url": "https://api.github.com/teams/1/members{/member}",
            "repositories_url": "https://api.github.com/teams/1/repos",
            "parent": None,
        }
        self.client.post.return_value = response

        team = await self.api.update(
            "foo",
            "bar",
            name="baz",
            description="A description",
            privacy=TeamPrivacy.CLOSED,
            parent_team_id="123",
        )

        self.client.post.assert_awaited_once_with(
            "/orgs/foo/teams/bar",
            data={
                "name": "baz",
                "description": "A description",
                "privacy": "closed",
                "parent_team_id": "123",
            },
        )

        self.assertEqual(team.id, 1)

    async def test_update_failure(self):
        response = create_response()
        self.client.post.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.update(
                "foo",
                "bar",
                name="baz",
                description="A description",
                privacy=TeamPrivacy.CLOSED,
                parent_team_id="123",
            )

        self.client.post.assert_awaited_once_with(
            "/orgs/foo/teams/bar",
            data={
                "name": "baz",
                "description": "A description",
                "privacy": "closed",
                "parent_team_id": "123",
            },
        )

    async def test_delete(self):
        response = create_response()
        self.client.delete.return_value = response

        await self.api.delete("foo", "bar")

        self.client.delete.assert_awaited_once_with(
            "/orgs/foo/teams/bar",
        )

    async def test_delete_failure(self):
        response = create_response()
        self.client.delete.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.delete("foo", "bar")

        self.client.delete.assert_awaited_once_with(
            "/orgs/foo/teams/bar",
        )

    async def test_members(self):
        response1 = create_response()
        response1.json.return_value = [
            {
                "id": 1,
                "login": "octocat",
                "node_id": "MDQ6VXNlcjE=",
                "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                "gravatar_id": "",
                "url": "https://api.github.com/users/octocat",
                "html_url": "https://github.com/octocat",
                "followers_url": "https://api.github.com/users/octocat/followers",
                "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                "organizations_url": "https://api.github.com/users/octocat/orgs",
                "repos_url": "https://api.github.com/users/octocat/repos",
                "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                "received_events_url": "https://api.github.com/users/octocat/received_events",
                "type": "User",
                "site_admin": False,
            }
        ]
        response2 = create_response()
        response2.json.return_value = [
            {
                "id": 2,
                "login": "octocat2",
                "node_id": "MDQ6VXNlcjE=",
                "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                "gravatar_id": "",
                "url": "https://api.github.com/users/octocat",
                "html_url": "https://github.com/octocat",
                "followers_url": "https://api.github.com/users/octocat/followers",
                "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                "organizations_url": "https://api.github.com/users/octocat/orgs",
                "repos_url": "https://api.github.com/users/octocat/repos",
                "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                "received_events_url": "https://api.github.com/users/octocat/received_events",
                "type": "User",
                "site_admin": False,
            },
            {
                "id": 3,
                "login": "octocat3",
                "node_id": "MDQ6VXNlcjE=",
                "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                "gravatar_id": "",
                "url": "https://api.github.com/users/octocat",
                "html_url": "https://github.com/octocat",
                "followers_url": "https://api.github.com/users/octocat/followers",
                "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                "organizations_url": "https://api.github.com/users/octocat/orgs",
                "repos_url": "https://api.github.com/users/octocat/repos",
                "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                "received_events_url": "https://api.github.com/users/octocat/received_events",
                "type": "User",
                "site_admin": False,
            },
        ]

        self.client.get_all.return_value = AsyncIteratorMock(
            [response1, response2]
        )

        async_it = aiter(self.api.members("foo", "bar"))
        member = await anext(async_it)
        self.assertEqual(member.id, 1)
        member = await anext(async_it)
        self.assertEqual(member.id, 2)
        member = await anext(async_it)
        self.assertEqual(member.id, 3)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/orgs/foo/teams/bar/members",
            params={"per_page": "100"},
        )

    async def test_update_member(self):
        response = create_response()
        self.client.put.return_value = response

        await self.api.update_member(
            "foo", "bar", "baz", role=TeamRole.MAINTAINER
        )

        self.client.put.assert_awaited_once_with(
            "/orgs/foo/teams/bar/memberships/baz", data={"role": "maintainer"}
        )

    async def test_add_member(self):
        response = create_response()
        self.client.put.return_value = response

        await self.api.add_member("foo", "bar", "baz", role=TeamRole.MAINTAINER)

        self.client.put.assert_awaited_once_with(
            "/orgs/foo/teams/bar/memberships/baz", data={"role": "maintainer"}
        )

    async def test_update_member_failure(self):
        response = create_response()
        self.client.put.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.update_member("foo", "bar", "baz")

        self.client.put.assert_awaited_once_with(
            "/orgs/foo/teams/bar/memberships/baz", data={"role": "member"}
        )

    async def test_remove_member(self):
        response = create_response()
        self.client.delete.return_value = response

        await self.api.remove_member("foo", "bar", "baz")

        self.client.delete.assert_awaited_once_with(
            "/orgs/foo/teams/bar/memberships/baz"
        )

    async def test_remove_member_failure(self):
        response = create_response()
        self.client.delete.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.remove_member("foo", "bar", "baz")

        self.client.delete.assert_awaited_once_with(
            "/orgs/foo/teams/bar/memberships/baz"
        )

    async def test_repositories(self):
        response1 = create_response()
        response1.json.return_value = [
            {
                "id": 1,
                "node_id": "MDEwOlJlcG9zaXRvcnkxMjk2MjY5",
                "name": "Hello-World",
                "full_name": "octocat/Hello-World",
                "owner": {
                    "login": "octocat",
                    "id": 1,
                    "node_id": "MDQ6VXNlcjE=",
                    "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                    "gravatar_id": "",
                    "url": "https://api.github.com/users/octocat",
                    "html_url": "https://github.com/octocat",
                    "followers_url": "https://api.github.com/users/octocat/followers",
                    "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                    "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                    "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                    "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                    "organizations_url": "https://api.github.com/users/octocat/orgs",
                    "repos_url": "https://api.github.com/users/octocat/repos",
                    "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                    "received_events_url": "https://api.github.com/users/octocat/received_events",
                    "type": "User",
                    "site_admin": False,
                },
                "private": False,
                "html_url": "https://github.com/octocat/Hello-World",
                "description": "This your first repo!",
                "fork": False,
                "url": "https://api.github.com/repos/octocat/Hello-World",
                "archive_url": "https://api.github.com/repos/octocat/Hello-World/{archive_format}{/ref}",
                "assignees_url": "https://api.github.com/repos/octocat/Hello-World/assignees{/user}",
                "blobs_url": "https://api.github.com/repos/octocat/Hello-World/git/blobs{/sha}",
                "branches_url": "https://api.github.com/repos/octocat/Hello-World/branches{/branch}",
                "collaborators_url": "https://api.github.com/repos/octocat/Hello-World/collaborators{/collaborator}",
                "comments_url": "https://api.github.com/repos/octocat/Hello-World/comments{/number}",
                "commits_url": "https://api.github.com/repos/octocat/Hello-World/commits{/sha}",
                "compare_url": "https://api.github.com/repos/octocat/Hello-World/compare/{base}...{head}",
                "contents_url": "https://api.github.com/repos/octocat/Hello-World/contents/{+path}",
                "contributors_url": "https://api.github.com/repos/octocat/Hello-World/contributors",
                "deployments_url": "https://api.github.com/repos/octocat/Hello-World/deployments",
                "downloads_url": "https://api.github.com/repos/octocat/Hello-World/downloads",
                "events_url": "https://api.github.com/repos/octocat/Hello-World/events",
                "forks_url": "https://api.github.com/repos/octocat/Hello-World/forks",
                "git_commits_url": "https://api.github.com/repos/octocat/Hello-World/git/commits{/sha}",
                "git_refs_url": "https://api.github.com/repos/octocat/Hello-World/git/refs{/sha}",
                "git_tags_url": "https://api.github.com/repos/octocat/Hello-World/git/tags{/sha}",
                "git_url": "git:github.com/octocat/Hello-World.git",
                "issue_comment_url": "https://api.github.com/repos/octocat/Hello-World/issues/comments{/number}",
                "issue_events_url": "https://api.github.com/repos/octocat/Hello-World/issues/events{/number}",
                "issues_url": "https://api.github.com/repos/octocat/Hello-World/issues{/number}",
                "keys_url": "https://api.github.com/repos/octocat/Hello-World/keys{/key_id}",
                "labels_url": "https://api.github.com/repos/octocat/Hello-World/labels{/name}",
                "languages_url": "https://api.github.com/repos/octocat/Hello-World/languages",
                "merges_url": "https://api.github.com/repos/octocat/Hello-World/merges",
                "milestones_url": "https://api.github.com/repos/octocat/Hello-World/milestones{/number}",
                "notifications_url": "https://api.github.com/repos/octocat/Hello-World/notifications{?since,all,participating}",
                "pulls_url": "https://api.github.com/repos/octocat/Hello-World/pulls{/number}",
                "releases_url": "https://api.github.com/repos/octocat/Hello-World/releases{/id}",
                "ssh_url": "git@github.com:octocat/Hello-World.git",
                "stargazers_url": "https://api.github.com/repos/octocat/Hello-World/stargazers",
                "statuses_url": "https://api.github.com/repos/octocat/Hello-World/statuses/{sha}",
                "subscribers_url": "https://api.github.com/repos/octocat/Hello-World/subscribers",
                "subscription_url": "https://api.github.com/repos/octocat/Hello-World/subscription",
                "tags_url": "https://api.github.com/repos/octocat/Hello-World/tags",
                "teams_url": "https://api.github.com/repos/octocat/Hello-World/teams",
                "trees_url": "https://api.github.com/repos/octocat/Hello-World/git/trees{/sha}",
                "clone_url": "https://github.com/octocat/Hello-World.git",
                "mirror_url": "git:git.example.com/octocat/Hello-World",
                "hooks_url": "https://api.github.com/repos/octocat/Hello-World/hooks",
                "svn_url": "https://svn.github.com/octocat/Hello-World",
                "homepage": "https://github.com",
                "language": None,
                "forks_count": 9,
                "stargazers_count": 80,
                "watchers_count": 80,
                "size": 108,
                "default_branch": "master",
                "open_issues_count": 0,
                "is_template": False,
                "topics": ["octocat", "atom", "electron", "api"],
                "has_issues": True,
                "has_projects": True,
                "has_wiki": True,
                "has_pages": False,
                "has_downloads": True,
                "has_discussions": False,
                "archived": False,
                "disabled": False,
                "visibility": "public",
                "pushed_at": "2011-01-26T19:06:43Z",
                "created_at": "2011-01-26T19:01:12Z",
                "updated_at": "2011-01-26T19:14:43Z",
                "permissions": {"admin": False, "push": False, "pull": True},
            }
        ]
        response2 = create_response()
        response2.json.return_value = [
            {
                "id": 2,
                "node_id": "MDEwOlJlcG9zaXRvcnkxMjk2MjY5",
                "name": "Hello-World",
                "full_name": "octocat/Hello-World",
                "owner": {
                    "login": "octocat",
                    "id": 1,
                    "node_id": "MDQ6VXNlcjE=",
                    "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                    "gravatar_id": "",
                    "url": "https://api.github.com/users/octocat",
                    "html_url": "https://github.com/octocat",
                    "followers_url": "https://api.github.com/users/octocat/followers",
                    "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                    "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                    "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                    "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                    "organizations_url": "https://api.github.com/users/octocat/orgs",
                    "repos_url": "https://api.github.com/users/octocat/repos",
                    "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                    "received_events_url": "https://api.github.com/users/octocat/received_events",
                    "type": "User",
                    "site_admin": False,
                },
                "private": False,
                "html_url": "https://github.com/octocat/Hello-World",
                "description": "This your first repo!",
                "fork": False,
                "url": "https://api.github.com/repos/octocat/Hello-World",
                "archive_url": "https://api.github.com/repos/octocat/Hello-World/{archive_format}{/ref}",
                "assignees_url": "https://api.github.com/repos/octocat/Hello-World/assignees{/user}",
                "blobs_url": "https://api.github.com/repos/octocat/Hello-World/git/blobs{/sha}",
                "branches_url": "https://api.github.com/repos/octocat/Hello-World/branches{/branch}",
                "collaborators_url": "https://api.github.com/repos/octocat/Hello-World/collaborators{/collaborator}",
                "comments_url": "https://api.github.com/repos/octocat/Hello-World/comments{/number}",
                "commits_url": "https://api.github.com/repos/octocat/Hello-World/commits{/sha}",
                "compare_url": "https://api.github.com/repos/octocat/Hello-World/compare/{base}...{head}",
                "contents_url": "https://api.github.com/repos/octocat/Hello-World/contents/{+path}",
                "contributors_url": "https://api.github.com/repos/octocat/Hello-World/contributors",
                "deployments_url": "https://api.github.com/repos/octocat/Hello-World/deployments",
                "downloads_url": "https://api.github.com/repos/octocat/Hello-World/downloads",
                "events_url": "https://api.github.com/repos/octocat/Hello-World/events",
                "forks_url": "https://api.github.com/repos/octocat/Hello-World/forks",
                "git_commits_url": "https://api.github.com/repos/octocat/Hello-World/git/commits{/sha}",
                "git_refs_url": "https://api.github.com/repos/octocat/Hello-World/git/refs{/sha}",
                "git_tags_url": "https://api.github.com/repos/octocat/Hello-World/git/tags{/sha}",
                "git_url": "git:github.com/octocat/Hello-World.git",
                "issue_comment_url": "https://api.github.com/repos/octocat/Hello-World/issues/comments{/number}",
                "issue_events_url": "https://api.github.com/repos/octocat/Hello-World/issues/events{/number}",
                "issues_url": "https://api.github.com/repos/octocat/Hello-World/issues{/number}",
                "keys_url": "https://api.github.com/repos/octocat/Hello-World/keys{/key_id}",
                "labels_url": "https://api.github.com/repos/octocat/Hello-World/labels{/name}",
                "languages_url": "https://api.github.com/repos/octocat/Hello-World/languages",
                "merges_url": "https://api.github.com/repos/octocat/Hello-World/merges",
                "milestones_url": "https://api.github.com/repos/octocat/Hello-World/milestones{/number}",
                "notifications_url": "https://api.github.com/repos/octocat/Hello-World/notifications{?since,all,participating}",
                "pulls_url": "https://api.github.com/repos/octocat/Hello-World/pulls{/number}",
                "releases_url": "https://api.github.com/repos/octocat/Hello-World/releases{/id}",
                "ssh_url": "git@github.com:octocat/Hello-World.git",
                "stargazers_url": "https://api.github.com/repos/octocat/Hello-World/stargazers",
                "statuses_url": "https://api.github.com/repos/octocat/Hello-World/statuses/{sha}",
                "subscribers_url": "https://api.github.com/repos/octocat/Hello-World/subscribers",
                "subscription_url": "https://api.github.com/repos/octocat/Hello-World/subscription",
                "tags_url": "https://api.github.com/repos/octocat/Hello-World/tags",
                "teams_url": "https://api.github.com/repos/octocat/Hello-World/teams",
                "trees_url": "https://api.github.com/repos/octocat/Hello-World/git/trees{/sha}",
                "clone_url": "https://github.com/octocat/Hello-World.git",
                "mirror_url": "git:git.example.com/octocat/Hello-World",
                "hooks_url": "https://api.github.com/repos/octocat/Hello-World/hooks",
                "svn_url": "https://svn.github.com/octocat/Hello-World",
                "homepage": "https://github.com",
                "language": None,
                "forks_count": 9,
                "stargazers_count": 80,
                "watchers_count": 80,
                "size": 108,
                "default_branch": "master",
                "open_issues_count": 0,
                "is_template": False,
                "topics": ["octocat", "atom", "electron", "api"],
                "has_issues": True,
                "has_projects": True,
                "has_wiki": True,
                "has_pages": False,
                "has_downloads": True,
                "has_discussions": False,
                "archived": False,
                "disabled": False,
                "visibility": "public",
                "pushed_at": "2011-01-26T19:06:43Z",
                "created_at": "2011-01-26T19:01:12Z",
                "updated_at": "2011-01-26T19:14:43Z",
                "permissions": {"admin": False, "push": False, "pull": True},
            },
            {
                "id": 3,
                "node_id": "MDEwOlJlcG9zaXRvcnkxMjk2MjY5",
                "name": "Hello-World",
                "full_name": "octocat/Hello-World",
                "owner": {
                    "login": "octocat",
                    "id": 1,
                    "node_id": "MDQ6VXNlcjE=",
                    "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                    "gravatar_id": "",
                    "url": "https://api.github.com/users/octocat",
                    "html_url": "https://github.com/octocat",
                    "followers_url": "https://api.github.com/users/octocat/followers",
                    "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                    "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                    "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                    "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                    "organizations_url": "https://api.github.com/users/octocat/orgs",
                    "repos_url": "https://api.github.com/users/octocat/repos",
                    "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                    "received_events_url": "https://api.github.com/users/octocat/received_events",
                    "type": "User",
                    "site_admin": False,
                },
                "private": False,
                "html_url": "https://github.com/octocat/Hello-World",
                "description": "This your first repo!",
                "fork": False,
                "url": "https://api.github.com/repos/octocat/Hello-World",
                "archive_url": "https://api.github.com/repos/octocat/Hello-World/{archive_format}{/ref}",
                "assignees_url": "https://api.github.com/repos/octocat/Hello-World/assignees{/user}",
                "blobs_url": "https://api.github.com/repos/octocat/Hello-World/git/blobs{/sha}",
                "branches_url": "https://api.github.com/repos/octocat/Hello-World/branches{/branch}",
                "collaborators_url": "https://api.github.com/repos/octocat/Hello-World/collaborators{/collaborator}",
                "comments_url": "https://api.github.com/repos/octocat/Hello-World/comments{/number}",
                "commits_url": "https://api.github.com/repos/octocat/Hello-World/commits{/sha}",
                "compare_url": "https://api.github.com/repos/octocat/Hello-World/compare/{base}...{head}",
                "contents_url": "https://api.github.com/repos/octocat/Hello-World/contents/{+path}",
                "contributors_url": "https://api.github.com/repos/octocat/Hello-World/contributors",
                "deployments_url": "https://api.github.com/repos/octocat/Hello-World/deployments",
                "downloads_url": "https://api.github.com/repos/octocat/Hello-World/downloads",
                "events_url": "https://api.github.com/repos/octocat/Hello-World/events",
                "forks_url": "https://api.github.com/repos/octocat/Hello-World/forks",
                "git_commits_url": "https://api.github.com/repos/octocat/Hello-World/git/commits{/sha}",
                "git_refs_url": "https://api.github.com/repos/octocat/Hello-World/git/refs{/sha}",
                "git_tags_url": "https://api.github.com/repos/octocat/Hello-World/git/tags{/sha}",
                "git_url": "git:github.com/octocat/Hello-World.git",
                "issue_comment_url": "https://api.github.com/repos/octocat/Hello-World/issues/comments{/number}",
                "issue_events_url": "https://api.github.com/repos/octocat/Hello-World/issues/events{/number}",
                "issues_url": "https://api.github.com/repos/octocat/Hello-World/issues{/number}",
                "keys_url": "https://api.github.com/repos/octocat/Hello-World/keys{/key_id}",
                "labels_url": "https://api.github.com/repos/octocat/Hello-World/labels{/name}",
                "languages_url": "https://api.github.com/repos/octocat/Hello-World/languages",
                "merges_url": "https://api.github.com/repos/octocat/Hello-World/merges",
                "milestones_url": "https://api.github.com/repos/octocat/Hello-World/milestones{/number}",
                "notifications_url": "https://api.github.com/repos/octocat/Hello-World/notifications{?since,all,participating}",
                "pulls_url": "https://api.github.com/repos/octocat/Hello-World/pulls{/number}",
                "releases_url": "https://api.github.com/repos/octocat/Hello-World/releases{/id}",
                "ssh_url": "git@github.com:octocat/Hello-World.git",
                "stargazers_url": "https://api.github.com/repos/octocat/Hello-World/stargazers",
                "statuses_url": "https://api.github.com/repos/octocat/Hello-World/statuses/{sha}",
                "subscribers_url": "https://api.github.com/repos/octocat/Hello-World/subscribers",
                "subscription_url": "https://api.github.com/repos/octocat/Hello-World/subscription",
                "tags_url": "https://api.github.com/repos/octocat/Hello-World/tags",
                "teams_url": "https://api.github.com/repos/octocat/Hello-World/teams",
                "trees_url": "https://api.github.com/repos/octocat/Hello-World/git/trees{/sha}",
                "clone_url": "https://github.com/octocat/Hello-World.git",
                "mirror_url": "git:git.example.com/octocat/Hello-World",
                "hooks_url": "https://api.github.com/repos/octocat/Hello-World/hooks",
                "svn_url": "https://svn.github.com/octocat/Hello-World",
                "homepage": "https://github.com",
                "language": None,
                "forks_count": 9,
                "stargazers_count": 80,
                "watchers_count": 80,
                "size": 108,
                "default_branch": "master",
                "open_issues_count": 0,
                "is_template": False,
                "topics": ["octocat", "atom", "electron", "api"],
                "has_issues": True,
                "has_projects": True,
                "has_wiki": True,
                "has_pages": False,
                "has_downloads": True,
                "has_discussions": False,
                "archived": False,
                "disabled": False,
                "visibility": "public",
                "pushed_at": "2011-01-26T19:06:43Z",
                "created_at": "2011-01-26T19:01:12Z",
                "updated_at": "2011-01-26T19:14:43Z",
                "permissions": {"admin": False, "push": False, "pull": True},
            },
        ]

        self.client.get_all.return_value = AsyncIteratorMock(
            [response1, response2]
        )

        async_it = aiter(self.api.repositories("foo", "bar"))
        repo = await anext(async_it)
        self.assertEqual(repo.id, 1)
        repo = await anext(async_it)
        self.assertEqual(repo.id, 2)
        repo = await anext(async_it)
        self.assertEqual(repo.id, 3)

        with self.assertRaises(StopAsyncIteration):
            await anext(async_it)

        self.client.get_all.assert_called_once_with(
            "/orgs/foo/teams/bar/repos",
            params={"per_page": "100"},
        )

    async def test_update_permissions(self):
        response = create_response()
        self.client.put.return_value = response

        await self.api.update_permission("foo", "bar", "baz", Permission.TRIAGE)

        self.client.put.assert_awaited_once_with(
            "/orgs/foo/teams/bar/repos/foo/baz", data={"permission": "triage"}
        )

    async def test_add_permissions(self):
        response = create_response()
        self.client.put.return_value = response

        await self.api.add_permission("foo", "bar", "baz", Permission.ADMIN)

        self.client.put.assert_awaited_once_with(
            "/orgs/foo/teams/bar/repos/foo/baz", data={"permission": "admin"}
        )

    async def test_update_permissions_failure(self):
        response = create_response()
        self.client.put.side_effect = httpx.HTTPStatusError(
            "404", request=MagicMock(), response=response
        )

        with self.assertRaises(httpx.HTTPStatusError):
            await self.api.update_permission(
                "foo", "bar", "baz", Permission.TRIAGE
            )

        self.client.put.assert_awaited_once_with(
            "/orgs/foo/teams/bar/repos/foo/baz", data={"permission": "triage"}
        )



# SPDX-FileCopyrightText: 2020-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from argparse import Namespace
from dataclasses import dataclass
from enum import IntEnum, auto
from pathlib import Path
from typing import Literal, Optional, SupportsInt, Union

import httpx

from pontos.changelog.conventional_commits import ChangelogBuilder
from pontos.errors import PontosError
from pontos.git import Git, ResetMode
from pontos.github.actions.core import ActionIO
from pontos.github.api import GitHubAsyncRESTApi
from pontos.release.command import AsyncCommand
from pontos.terminal import Terminal
from pontos.version import Version, VersionError
from pontos.version.helper import get_last_release_version
from pontos.version.project import Project
from pontos.version.schemes import VersioningScheme

from .helper import (
    ReleaseType,
    find_signing_key,
    get_next_release_version,
    repository_split,
)


@dataclass
class ReleaseInformation:
    last_release_version: Optional[Version]
    release_version: Version
    git_release_tag: str
    next_version: Optional[Version]

    def write_github_output(self):
        with ActionIO.out() as output:
            output.write(
                "last-release-version", self.last_release_version or ""
            )
            output.write("release-version", self.release_version)
            output.write("git-release-tag", self.git_release_tag)
            output.write("next-version", self.next_version or "")


class CreateReleaseReturnValue(IntEnum):
    """
    Possible return values of ReleaseCommand
    """

    SUCCESS = 0
    PROJECT_SETTINGS_NOT_FOUND = auto()
    TOKEN_MISSING = auto()
    NO_LAST_RELEASE_VERSION = auto()
    NO_RELEASE_VERSION = auto()
    ALREADY_TAKEN = auto()
    CREATE_RELEASE_ERROR = auto()
    UPDATE_VERSION_ERROR = auto()
    UPDATE_VERSION_AFTER_RELEASE_ERROR = auto()
    INVALID_REPOSITORY = auto()


class CreateReleaseCommand(AsyncCommand):
    """
    A CLI command for creating a release

    Args:
        terminal: A Terminal for output
    """

    def __init__(self, *, terminal: Terminal, error_terminal: Terminal) -> None:
        super().__init__(terminal=terminal, error_terminal=error_terminal)
        self.git = Git()

    def _has_tag(self, git_version: str) -> bool:
        git_tags = self.git.list_tags()
        return git_version in git_tags

    def _create_changelog(
        self,
        release_version: Version,
        last_release_version: Optional[Version],
        cc_config: Optional[Path],
    ) -> str:
        changelog_builder = ChangelogBuilder(
            repository=self.repository,
            config=cc_config,
            git_tag_prefix=self.git_tag_prefix,
        )

        return changelog_builder.create_changelog(
            last_version=(
                last_release_version.parsed_version
                if last_release_version
                else None
            ),
            next_version=release_version,
        )

    async def _create_release(
        self,
        release_version: Version,
        token: str,
        release_text: str,
        github_pre_release: bool,
    ) -> None:
        github = GitHubAsyncRESTApi(token=token)

        git_version = f"{self.git_tag_prefix}{release_version}"
        _, project = repository_split(self.repository)

        await github.releases.create(
            self.repository,
            git_version,
            name=f"{project} {release_version}",
            body=release_text,
            prerelease=release_version.is_pre_release or github_pre_release,
        )

    async def async_run(  # type: ignore[override]
        self,
        *,
        token: str,
        repository: str,
        versioning_scheme: VersioningScheme,
        release_type: ReleaseType,
        release_version: Optional[Version],
        next_version: Union[Version, Literal[False], None],
        git_signing_key: str,
        git_remote_name: Optional[str],
        git_tag_prefix: Optional[str],
        cc_config: Optional[Path],
        local: Optional[bool] = False,
        release_series: Optional[str] = None,
        update_project: bool = True,
        github_pre_release: bool = False,
    ) -> CreateReleaseReturnValue:
        """
        Create a release

        Args:
            token: A token for creating a release on GitHub
            repository: GitHub repository (owner/name)
            versioning_scheme: The versioning scheme to use for version parsing
                and calculation
            release_type: Type of the release to prepare. Defines the release
                version. PATCH increments the bugfix version. CALENDAR creates
                a new CalVer release version. VERSION uses the provided
                release_version.
            release_version: Optional release version to use. If not set the
                to be released version will be determined from the project.
            next_version: Optional version to set after the release.
                If set to None the next development version will be set.
                If set to False the version will not be changed after the
                release. Default is to update to the next development version.
            git_signing_key: A GPG key ID to use for creating signatures.
            git_remote_name: Name of the git remote to use.
            git_tag_prefix: An optional prefix to use for creating a git tag
                from the release version.
            cc_config: A path to a settings file for creating conventional
                commits.
            local: Only create changes locally and don't push changes to
                remote repository. Also don't create a GitHub release.
            release_series: Optional release series to use.
                For example: "1.2", "2", "23".
            update_project: Update version in project files.
            github_pre_release: Enforce uploading a release as a GitHub pre
                release
        """
        git_signing_key = (
            git_signing_key
            if git_signing_key is not None
            else find_signing_key(self.terminal)
        )
        self.git_tag_prefix = git_tag_prefix or ""
        self.repository = repository

        try:
            _, self.project_name = repository_split(repository)
        except ValueError as e:
            self.print_error(str(e))
            return CreateReleaseReturnValue.INVALID_REPOSITORY

        self.terminal.info(f"Using versioning scheme {versioning_scheme.name}")

        try:
            last_release_version = get_last_release_version(
                parse_version=versioning_scheme.parse_version,
                git_tag_prefix=self.git_tag_prefix,
                tag_name=(
                    f"{self.git_tag_prefix}{release_series}.*"
                    if release_series
                    else None
                ),
                # include changes from pre-releases in release changelog for
                # non pre-release changes
                ignore_pre_releases=release_type
                not in [
                    ReleaseType.ALPHA,
                    ReleaseType.BETA,
                    ReleaseType.RELEASE_CANDIDATE,
                ]
                # but not when using a release series because then we might not
                # be able to determine the last release if there are only
                # pre-releases in the series yet
                and not release_series,
            )
        except PontosError as e:
            last_release_version = None
            self.print_warning(f"Could not determine last release version. {e}")

        if not last_release_version:
            if not release_version:
                self.print_error("Unable to determine last release version.")
                return CreateReleaseReturnValue.NO_LAST_RELEASE_VERSION
            else:
                self.terminal.info(
                    f"Creating the initial release {release_version}"
                )

        else:
            self.terminal.info(f"Last release was {last_release_version}")

        calculator = versioning_scheme.calculator()

        try:
            release_version = get_next_release_version(
                last_release_version=last_release_version,
                calculator=calculator,
                release_type=release_type,
                release_version=release_version,
            )
        except VersionError as e:
            self.print_error(f"Unable to determine release version. {e}")
            return CreateReleaseReturnValue.NO_RELEASE_VERSION

        self.terminal.info(f"Preparing the release {release_version}")

        git_version = f"{self.git_tag_prefix}{release_version}"

        if self._has_tag(git_version):
            self.print_error(f"Git tag {git_version} already exists.")
            return CreateReleaseReturnValue.ALREADY_TAKEN

        if update_project:
            try:
                project = Project(versioning_scheme)
            except PontosError as e:
                self.print_error(f"Unable to determine project settings. {e}")
                return CreateReleaseReturnValue.PROJECT_SETTINGS_NOT_FOUND

            try:
                updated = project.update_version(release_version)

                self.terminal.ok(f"Updated version to {release_version}")

                for path in updated.changed_files:
                    self.terminal.info(f"Adding changes of {path}")
                    self.git.add(path)
            except VersionError as e:
                self.terminal.error(
                    f"Unable to update version to {release_version}. {e}"
                )
                return CreateReleaseReturnValue.UPDATE_VERSION_ERROR

        if last_release_version:
            self.terminal.info(
                f"Creating changelog for {release_version} since "
                f"{last_release_version}"
            )
        else:
            self.terminal.info(
                f"Creating changelog for {release_version} as initial release."
            )

        release_text = self._create_changelog(
            release_version, last_release_version, cc_config
        )

        commit_msg = f"Automatic release to {release_version}"

        # check if files have been modified and create a commit
        status = list(self.git.status())
        if status:
            self.terminal.info("Committing changes")

            self.git.commit(
                commit_msg, verify=False, gpg_signing_key=git_signing_key
            )

        self.terminal.info(f"Creating tag {git_version}")
        self.git.tag(
            git_version, gpg_key_id=git_signing_key, message=commit_msg
        )

        if not local:
            self.terminal.info("Pushing changes")
            self.git.push(follow_tags=True, remote=git_remote_name)

            try:
                self.terminal.info(f"Creating release for {release_version}")

                await self._create_release(
                    release_version,
                    token,
                    release_text,
                    github_pre_release,
                )

                self.terminal.ok(f"Created release {release_version}")
            except httpx.HTTPStatusError as e:
                self.print_error(str(e))
                # revert commit and tag
                self.git.delete_tag(git_version)
                self.git.push(git_version, delete=True, remote=git_remote_name)
                self.git.reset("HEAD^", mode=ResetMode.HARD)
                self.git.push(force=True, remote=git_remote_name)
                return CreateReleaseReturnValue.CREATE_RELEASE_ERROR

        if next_version is None:
            next_version = calculator.next_dev_version(release_version)

        if next_version:
            if update_project:
                try:
                    updated = project.update_version(next_version)
                    self.terminal.ok(
                        f"Updated version after release to {next_version}"
                    )
                except VersionError as e:
                    self.print_error(
                        f"Error while updating version after release. {e}"
                    )
                    return (
                        CreateReleaseReturnValue.UPDATE_VERSION_AFTER_RELEASE_ERROR
                    )

                for f in updated.changed_files:
                    self.terminal.info(f"Adding changes of {f}")
                    self.git.add(f)

            # check if files have been modified and create a commit
            status = list(self.git.status())
            if status:
                commit_msg = f"""Automatic adjustments after release [skip ci]

* Update to version {next_version}
"""

                self.terminal.info("Committing changes after release")
                self.git.commit(
                    commit_msg, verify=False, gpg_signing_key=git_signing_key
                )

            if not local:
                self.terminal.info("Pushing changes")
                self.git.push(follow_tags=True, remote=git_remote_name)

        self.release_information = ReleaseInformation(
            last_release_version=last_release_version,
            release_version=release_version,
            git_release_tag=git_version,
            next_version=next_version or None,
        )
        if ActionIO.has_output():
            self.release_information.write_github_output()

        return CreateReleaseReturnValue.SUCCESS


def create_release(
    args: Namespace,
    *,
    token: str,
    terminal: Terminal,
    error_terminal: Terminal,
    **_kwargs,
) -> SupportsInt:
    if not token:
        error_terminal.error(
            "Token is missing. The GitHub token is required to create a "
            "release."
        )
        return CreateReleaseReturnValue.TOKEN_MISSING

    return CreateReleaseCommand(
        terminal=terminal, error_terminal=error_terminal
    ).run(
        token=token,
        repository=args.repository,
        versioning_scheme=args.versioning_scheme,
        release_type=args.release_type,
        release_version=args.release_version,
        next_version=args.next_version,
        git_remote_name=args.git_remote_name,
        git_signing_key=args.git_signing_key,
        git_tag_prefix=args.git_tag_prefix,
        cc_config=args.cc_config,
        local=args.local,
        release_series=args.release_series,
        update_project=args.update_project,
        github_pre_release=args.github_pre_release,
    )



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from unittest.mock import MagicMock

import httpx

from pontos.github.api.client import GitHubAsyncREST, GitHubAsyncRESTClient
from tests import AsyncMock, IsolatedAsyncioTestCase


def create_response(*args, **kwargs) -> MagicMock:
    return MagicMock(spec=httpx.Response, *args, **kwargs)


class GitHubAsyncRESTTestCase(IsolatedAsyncioTestCase):
    api_cls = GitHubAsyncREST

    def setUp(self) -> None:
        self.client = AsyncMock(spec=GitHubAsyncRESTClient)
        self.api = self.api_cls(self.client)



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import asyncio
from argparse import Namespace
from typing import Callable

import httpx

from pontos.nvd.cve.api import CVEApi

from ._parser import cve_parser, cves_parser

__all__ = ("CVEApi",)


async def query_cves(args: Namespace) -> None:
    async with CVEApi(token=args.token) as api:
        async for cve in api.cves(
            keywords=args.keywords,
            cpe_name=args.cpe_name,
            cvss_v2_vector=args.cvss_v2_vector,
            cvss_v3_vector=args.cvss_v3_vector,
            source_identifier=args.source_identifier,
            request_results=args.number,
            start_index=args.start,
        ):
            print(cve)


async def query_cve(args: Namespace) -> None:
    async with CVEApi(token=args.token) as api:
        cve = await api.cve(args.cve_id)
        print(cve)


def cves_main() -> None:
    main(cves_parser(), query_cves)


def cve_main() -> None:
    main(cve_parser(), query_cve)


def main(args: Namespace, func: Callable) -> None:
    try:
        asyncio.run(func(args))
    except KeyboardInterrupt:
        pass
    except httpx.HTTPStatusError as e:
        print(f"HTTP Error {e.response.status_code}: {e.response.text}")



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional
from uuid import UUID

from pontos.models import Model


@dataclass
class CPEMatch(Model):
    """
    Represents a single CPE match.

    Attributes:
        cpe_name: Name of the matching CPE
        cpe_name_id: Name ID of the matching CPE
    """

    cpe_name: str
    cpe_name_id: UUID


@dataclass
class CPEMatchString(Model):
    """
    Represents a CPE match string, matching criteria to one or more CPEs

    Attributes:
        match_criteria_id: The identifier of the CPE match
        criteria: The CPE formatted match criteria
        version_start_including: Optional start of the matching version range, including the given version
        version_start_excluding: Optional start of the matching version range, excluding the given version
        version_end_including: Optional end of the matching version range, including the given version
        version_end_excluding: Optional end of the matching version range, excluding the given version
        status: Status of the CPE match
        cpe_last_modified: Optional date the CPEs list of the match was last modified
        created: Creation date of the CPE
        last_modified: Last modification date of the CPE
        matches: List of CPEs matching the criteria string and the optional range limits
    """

    match_criteria_id: UUID
    criteria: str
    status: str
    created: datetime
    last_modified: datetime
    cpe_last_modified: Optional[datetime] = None
    matches: List[CPEMatch] = field(default_factory=list)
    version_start_including: Optional[str] = None
    version_start_excluding: Optional[str] = None
    version_end_including: Optional[str] = None
    version_end_excluding: Optional[str] = None



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
This script locks a branch in a repo of an organization via branch protection
"""

from argparse import ArgumentParser, BooleanOptionalAction, Namespace

from pontos.github.api import GitHubAsyncRESTApi
from pontos.github.api.branch import update_from_applied_settings


def add_script_arguments(parser: ArgumentParser) -> None:
    parser.add_argument("repository", help="org/repo combination")
    parser.add_argument("branch", help="branch to use")
    parser.add_argument(
        "--allow",
        action=BooleanOptionalAction,
        help="Allow/disallow admin users to bypass the branch protection rules",
    )


async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:
    branch_protection = await api.branches.protection_rules(
        args.repository, args.branch
    )
    if args.allow:
        enforce_admins = False
    else:
        enforce_admins = True

    kwargs = update_from_applied_settings(
        branch_protection, enforce_admins=enforce_admins
    )
    await api.branches.update_protection_rules(
        args.repository,
        args.branch,
        **kwargs,
    )
    if args.allow:
        print(
            f"Allowed admins to bypass the branch protection rules for branch "
            f"{args.branch} in {args.repository} now."
        )
    else:
        print(
            f"Admin users are not allowed to bypass the branch protection "
            f"rules for branch {args.branch} in {args.repository} now."
        )
    return 0



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

"""
This script prints out all repositories existing in the space of the given
organization
"""

from argparse import ArgumentParser, Namespace

from pontos.github.api import GitHubAsyncRESTApi


def add_script_arguments(parser: ArgumentParser) -> None:
    parser.add_argument("organization", help="owner")
    parser.add_argument("repository", nargs="*", help="repo")
    parser.add_argument("--topics", nargs="*", help="new topics to set")


async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:
    for repository in args.repository:
        topics = set(
            await api.repositories.topics(f"{args.organization}/{repository}")
        )
        topics = topics.union(args.topics)
        print(topics)
        topics = await api.repositories.update_topics(
            f"{args.organization}/{repository}", topics
        )

        print(f"{args.organization}/{repository}: ", end="")
        for topic in topics:
            print(f"{topic} ", end="")
        print()

    return 0



# Copyright (C) 2022 - 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# ruff: noqa: F403

from pontos.github.models.artifact import *
from pontos.github.models.base import *
from pontos.github.models.branch import *
from pontos.github.models.organization import *
from pontos.github.models.packages import *
from pontos.github.models.pull_request import *
from pontos.github.models.release import *
from pontos.github.models.search import *
from pontos.github.models.tag import *
from pontos.github.models.workflow import *



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from dataclasses import dataclass
from typing import Optional

from pontos.models import Model, StrEnum


class Severity(StrEnum):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


class AccessVector(StrEnum):
    NETWORK = "NETWORK"
    ADJACENT_NETWORK = "ADJACENT_NETWORK"
    LOCAL = "LOCAL"


class AccessComplexity(StrEnum):
    HIGH = "HIGH"
    MEDIUM = "MEDIUM"
    LOW = "LOW"


class Authentication(StrEnum):
    MULTIPLE = "MULTIPLE"
    SINGLE = "SINGLE"
    NONE = "NONE"


class Impact(StrEnum):
    NONE = "NONE"
    PARTIAL = "PARTIAL"
    COMPLETE = "COMPLETE"


class Exploitability(StrEnum):
    UNPROVEN = "UNPROVEN"
    PROOF_OF_CONCEPT = "PROOF_OF_CONCEPT"
    FUNCTIONAL = "FUNCTIONAL"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class RemediationLevel(StrEnum):
    OFFICIAL_FIX = "OFFICIAL_FIX"
    TEMPORARY_FIX = "TEMPORARY_FIX"
    WORKAROUND = "WORKAROUND"
    UNAVAILABLE = "UNAVAILABLE"
    NOT_DEFINED = "NOT_DEFINED"


class ReportConfidence(StrEnum):
    UNCONFIRMED = "UNCONFIRMED"
    UNCORROBORATED = "UNCORROBORATED"
    CONFIRMED = "CONFIRMED"
    NOT_DEFINED = "NOT_DEFINED"


class CollateralDamagePotential(StrEnum):
    NONE = "NONE"
    LOW = "LOW"
    LOW_MEDIUM = "LOW_MEDIUM"
    MEDIUM_HIGH = "MEDIUM_HIGH"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class TargetDistribution(StrEnum):
    NONE = "NONE"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class Requirement(StrEnum):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


@dataclass
class CVSSData(Model):
    version: str
    vector_string: str
    base_score: float
    access_vector: Optional[AccessVector] = None
    access_complexity: Optional[AccessComplexity] = None
    authentication: Optional[Authentication] = None
    confidentiality_impact: Optional[Impact] = None
    integrity_impact: Optional[Impact] = None
    availability_impact: Optional[Impact] = None
    exploitability: Optional[Exploitability] = None
    remediation_level: Optional[RemediationLevel] = None
    report_confidence: Optional[ReportConfidence] = None
    temporal_score: Optional[float] = None
    collateral_damage_potential: Optional[CollateralDamagePotential] = None
    target_distribution: Optional[TargetDistribution] = None
    confidentiality_requirement: Optional[Requirement] = None
    integrity_requirement: Optional[Requirement] = None
    availability_requirement: Optional[Requirement] = None
    environmental_score: Optional[float] = None



# SPDX-FileCopyrightText: 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

from pontos.github.api.client import GitHubAsyncREST
from pontos.github.models.billing import (
    ActionsBilling,
    PackagesBilling,
    StorageBilling,
)


class GitHubAsyncRESTBilling(GitHubAsyncREST):
    async def actions(self, organization: str) -> ActionsBilling:
        """
        Get the summary of the free and paid GitHub Actions minutes used

        https://docs.github.com/en/rest/billing/billing#get-github-actions-billing-for-an-organization

        Args:
            organization: The organization name

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the Actions billing

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    billing = await api.billing.actions("foo")
                    print(billing)
        """
        api = f"/orgs/{organization}/settings/billing/actions"
        response = await self._client.get(api)
        response.raise_for_status()
        return ActionsBilling.from_dict(response.json())

    async def packages(self, organization: str) -> PackagesBilling:
        """
        Get the free and paid storage used for GitHub Packages in gigabytes

        https://docs.github.com/en/rest/billing/billing#get-github-packages-billing-for-an-organization

        Args:
            organization: The organization name

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the Packages billing

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    billing = await api.billing.packages("foo")
                    print(billing)
        """
        api = f"/orgs/{organization}/settings/billing/packages"
        response = await self._client.get(api)
        response.raise_for_status()
        return PackagesBilling.from_dict(response.json())

    async def storage(self, organization: str) -> StorageBilling:
        """
        Get the estimated paid and estimated total storage used for GitHub
        Actions and GitHub Packages

        https://docs.github.com/en/rest/billing/billing#get-shared-storage-billing-for-an-organization

        Args:
            organization: The organization name

        Raises:
            HTTPStatusError: A httpx.HTTPStatusError is raised if the request
                failed.

        Returns:
            Information about the storage billing

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    billing = await api.billing.storage("foo")
                    print(billing)
        """
        api = f"/orgs/{organization}/settings/billing/shared-storage"
        response = await self._client.get(api)
        response.raise_for_status()
        return StorageBilling.from_dict(response.json())



# SPDX-FileCopyrightText: 2019-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from abc import ABC, abstractmethod
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from shutil import get_terminal_size
from typing import Any, Callable, Generator, Optional

import colorful as cf  # type: ignore

from pontos.helper import DownloadProgressIterable

TERMINAL_SIZE_FALLBACK = (80, 24)  # use a small standard size as fallback


class Signs(Enum):
    FAIL = "\N{HEAVY MULTIPLICATION X}"
    ERROR = "\N{MULTIPLICATION SIGN}"
    WARNING = "\N{WARNING SIGN}"
    OK = "\N{CHECK MARK}"
    INFO = "\N{INFORMATION SOURCE}"
    NONE = " "

    def __str__(self):
        return f"{self.value}"


STATUS_LEN = 2


class Terminal(ABC):
    """
    Abstract base class representing a terminal console
    """

    def __init__(self) -> None:
        super().__init__()
        self._indent = 0

    @contextmanager
    def indent(self, indentation: int = 4) -> Generator[None, None, None]:
        """
        A context manager for indenting output using spaces

        Example:
            .. code-block:: python

                with terminal.indent():
                    terminal.print("...")

        Args:
            indentation: Number of spaces to be used for indentation.
                By default 4.
        """
        current_indent = self._indent
        self._add_indent(indentation)

        yield

        self._indent = current_indent

    def _add_indent(self, indentation: int = 4) -> None:
        self._indent += indentation

    @abstractmethod
    def out(self, *messages: Any, **kwargs: Any) -> None:
        """
        Print messages without formatting.

        Args:
            *messages: Arguments to print.
            **kwargs: Keyword arguments forwarded to the underlying
                     implementation.
        """

    @abstractmethod
    def print(self, *messages: Any, **kwargs: Any) -> None:
        """
        Print messages. Possibly formatting is applied.

        Args:
            *messages: Arguments to print.
            **kwargs: Keyword arguments forwarded to the underlying
                     implementation.
        """

    @abstractmethod
    def ok(self, *messages: Any, **kwargs: Any) -> None:
        """
        Print a success message. Possibly formatting is applied.

        Args:
            *messages: Arguments to print.
            **kwargs: Keyword arguments forwarded to the underlying
                     implementation.
        """

    @abstractmethod
    def fail(self, *messages: Any, **kwargs: Any) -> None:
        """
        Print a failure message. Possibly formatting is applied.

        Args:
            *messages: Arguments to print.
            **kwargs: Keyword arguments forwarded to the underlying
                     implementation.
        """

    @abstractmethod
    def error(self, *messages: Any, **kwargs: Any) -> None:
        """
        Print an error message. Possibly formatting is applied.

        Args:
            *messages: Arguments to print.
            **kwargs: Keyword arguments forwarded to the underlying
                     implementation.
        """

    @abstractmethod
    def warning(self, *messages: Any, **kwargs: Any) -> None:
        """
        Print a warning message. Possibly formatting is applied.

        Args:
            *messages: Arguments to print.
            **kwargs: Keyword arguments forwarded to the underlying
                     implementation.
        """

    @abstractmethod
    def info(self, *messages: Any, **kwargs: Any) -> None:
        """
        Print an info message. Possibly formatting is applied.

        Args:
            *messages: Arguments to print.
            **kwargs: Keyword arguments forwarded to the underlying
                     implementation.
        """

    @abstractmethod
    def bold_info(self, *messages: Any, **kwargs: Any) -> None:
        """
        Print an info message with bold text. Possibly formatting is applied.

        Args:
            *messages: Arguments to print.
            **kwargs: Keyword arguments forwarded to the underlying
                     implementation.
        """

    @abstractmethod
    def download_progress(self, progress: DownloadProgressIterable) -> None:
        """
        Display a download progress
        """


class ConsoleTerminal(Terminal):
    """
    A simple Terminal using colorful internally for highlighting
    """

    # Keep arguments for backwards compatibility but ignore them
    # pylint: disable=unused-argument
    def __init__(self, *, verbose: int = 1, log_file: Optional[Path] = None):
        super().__init__()

    @staticmethod
    def get_width() -> int:
        """
        Get the width of the terminal window
        """
        width, _ = get_terminal_size(TERMINAL_SIZE_FALLBACK)
        return width

    def _print_status(
        self,
        *messages: Any,
        status: Signs,
        color: Callable,
        style: Callable = cf.reset,
        new_line: bool = True,
        **kwargs: Any,
    ) -> None:
        width = self.get_width()
        offset = self._indent + STATUS_LEN
        usable_width = width - offset

        # deal with existing newlines, to avoid breaking the formatting
        # done by the terminal
        message = "".join(messages)
        processed_messages = message.split("\n")
        output = self._format_message(
            message=processed_messages[0],
            usable_width=usable_width,
            offset=offset,
            first=True,
        )
        if len(processed_messages) > 0:
            for msg in processed_messages[1:]:
                output += "\n"
                output += self._format_message(
                    message=msg,
                    usable_width=usable_width,
                    offset=offset,
                )
        if new_line:
            print(style(f"{color(status)} {output}"), **kwargs)
        else:
            kwargs.update({"end": "", "flush": True})
            print(style(f"{color(status)} {output}"), **kwargs)

    def _format_message(
        self,
        message: str,
        usable_width: int,
        offset: int,
        *,
        first: bool = False,
    ) -> str:
        formatted_message = ""
        if first:
            formatted_message += " " * self._indent
        else:
            formatted_message += " " * offset

        while usable_width < len(message):
            part = message[:usable_width]
            message = message[usable_width:]
            formatted_message += f"{part}"

            if len(message) > 0:
                formatted_message += f'\n{" " * offset}'

        formatted_message += f"{message}"

        return formatted_message

    def out(self, *messages: Any, **kwargs: Any) -> None:
        self.print(*messages, **kwargs)

    def print(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.NONE, "color": cf.white})
        self._print_status(*messages, **kwargs)

    def ok(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.OK, "color": cf.green})
        self._print_status(*messages, **kwargs)

    def fail(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.FAIL, "color": cf.red})
        self._print_status(*messages, **kwargs)

    def error(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.ERROR, "color": cf.red})
        self._print_status(*messages, **kwargs)

    def warning(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.WARNING, "color": cf.yellow})
        self._print_status(*messages, **kwargs)

    def info(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update({"status": Signs.INFO, "color": cf.cyan})
        self._print_status(*messages, **kwargs)

    def bold_info(self, *messages: Any, **kwargs: Any) -> None:
        kwargs.update(
            {"status": Signs.INFO, "color": cf.cyan, "style": cf.bold}
        )
        self._print_status(*messages, **kwargs)

    def download_progress(self, progress: DownloadProgressIterable) -> None:
        spinner = ["-", "\\", "|", "/"]
        if progress.length:
            for percent in progress:
                done = int(50 * percent) if percent else 0
                self.out(
                    f"\r[{'=' * done}{' ' * (50-done)}]", end="", flush=True
                )
        else:
            i = 0
            for _ in progress:
                i = i + 1
                if i == 4:
                    i = 0
                self.out(f"\r[{spinner[i]}]", end="", flush=True)

        self.out(f"\r[{Signs.OK}]{' ' * 50}", end="", flush=True)



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import asyncio
from argparse import Namespace
from typing import Callable

import httpx

from pontos.nvd.cpe_match.api import CPEMatchApi

from ._parser import cpe_match_parse, cpe_matches_parse

__all__ = ("CPEMatchApi",)


async def query_cpe_match(args: Namespace) -> None:
    async with CPEMatchApi(token=args.token) as api:
        cpe_match = await api.cpe_match(args.match_criteria_id)
        print(cpe_match)


async def query_cpe_matches(args: Namespace) -> None:
    async with CPEMatchApi(token=args.token) as api:
        response = api.cpe_matches(
            cve_id=args.cve_id,
            request_results=args.number,
            start_index=args.start,
        )
        async for cpe_match in response:
            print(cpe_match)


def cpe_match_main() -> None:
    main(cpe_match_parse(), query_cpe_match)


def cpe_matches_main() -> None:
    main(cpe_matches_parse(), query_cpe_matches)


def main(args: Namespace, func: Callable) -> None:
    try:
        asyncio.run(func(args))
    except KeyboardInterrupt:
        pass
    except httpx.HTTPStatusError as e:
        print(f"HTTP Error {e.response.status_code}: {e.response.text}")



# pontos/release/release.py
# SPDX-FileCopyrightText: 2020-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import subprocess
import sys
from typing import NoReturn

from pontos.git import GitError
from pontos.terminal.null import NullTerminal
from pontos.terminal.rich import RichTerminal

from ._parser import parse_args


def main(
    args=None,
) -> NoReturn:
    username, token, parsed_args = parse_args(args)
    logging.basicConfig(format="%(levelname)s - %(name)s - %(message)s")

    if parsed_args.quiet:
        term = NullTerminal()
        error_terminal = NullTerminal()
        logging.disable()
    else:
        term = RichTerminal()  # type: ignore[assignment]
        error_terminal = RichTerminal(file=sys.stderr)  # type: ignore[assignment] # noqa: E501

    try:
        retval = parsed_args.func(
            parsed_args,
            terminal=term,
            error_terminal=error_terminal,
            username=username,
            token=token,
        )
        sys.exit(int(retval))
    except KeyboardInterrupt:
        sys.exit(1)
    except GitError as e:
        error_terminal.error(f'Could not run git command "{e.cmd}".')
        error = e.stderr if e.stderr else e.stdout
        error_terminal.print(f"Output was: {error}")
        sys.exit(1)
    except subprocess.CalledProcessError as e:
        if "--passphrase" not in e.cmd:
            error_terminal.error(f'Could not run command "{e.cmd}".')
        else:
            error_terminal.error("Headless signing failed.")

        error_terminal.print(f"Error was: {e.stderr}")
        sys.exit(1)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022-2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# pylint: disable=arguments-differ,redefined-builtin

from unittest.mock import MagicMock, call, patch

from pontos.github.api.client import (
    DEFAULT_ACCEPT_HEADER,
    GITHUB_API_VERSION,
    GitHubAsyncRESTClient,
)
from pontos.github.api.helper import DEFAULT_GITHUB_API_URL
from tests import AsyncMock, IsolatedAsyncioTestCase, aiter, anext


class GitHubAsyncRESTClientTestCase(IsolatedAsyncioTestCase):
    @patch("pontos.github.api.client.httpx.AsyncClient")
    def setUp(self, async_client: MagicMock) -> None:
        self.http_client = AsyncMock()
        async_client.return_value = self.http_client
        self.client = GitHubAsyncRESTClient("token")

    async def test_get(self):
        await self.client.get("/foo/bar")

        self.http_client.get.assert_awaited_once_with(
            f"{DEFAULT_GITHUB_API_URL}/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            params=None,
            follow_redirects=True,
        )

    async def test_get_url(self):
        await self.client.get("https://github.com/foo/bar")

        self.http_client.get.assert_awaited_once_with(
            "https://github.com/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            params=None,
            follow_redirects=True,
        )

    async def test_get_all(self):
        url = "https://foo.bar"
        response1 = MagicMock(links={"next": {"url": url}})
        response2 = MagicMock(links=None)

        self.http_client.get.side_effect = [
            response1,
            response2,
        ]
        it = aiter(self.client.get_all("/foo/bar"))

        await anext(it)
        await anext(it)

        with self.assertRaises(StopAsyncIteration):
            await anext(it)

        self.http_client.get.assert_has_awaits(
            [
                call(
                    f"{DEFAULT_GITHUB_API_URL}/foo/bar",
                    headers={
                        "Accept": DEFAULT_ACCEPT_HEADER,
                        "Authorization": "token token",
                        "X-GitHub-Api-Version": GITHUB_API_VERSION,
                    },
                    params=None,
                    follow_redirects=True,
                ),
                call(
                    f"{url}",
                    headers={
                        "Accept": DEFAULT_ACCEPT_HEADER,
                        "Authorization": "token token",
                        "X-GitHub-Api-Version": GITHUB_API_VERSION,
                    },
                    params=None,
                    follow_redirects=True,
                ),
            ]
        )

    async def test_delete(self):
        await self.client.delete("/foo/bar")

        self.http_client.delete.assert_awaited_once_with(
            f"{DEFAULT_GITHUB_API_URL}/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            params=None,
        )

    async def test_delete_url(self):
        await self.client.delete("https://github.com/foo/bar")

        self.http_client.delete.assert_awaited_once_with(
            "https://github.com/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            params=None,
        )

    async def test_post(self):
        await self.client.post("/foo/bar", data={"foo": "bar"})

        self.http_client.post.assert_awaited_once_with(
            f"{DEFAULT_GITHUB_API_URL}/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            json={"foo": "bar"},
            params=None,
            content=None,
        )

    async def test_post_url(self):
        await self.client.post(
            "https://github.com/foo/bar", data={"foo": "bar"}
        )

        self.http_client.post.assert_awaited_once_with(
            "https://github.com/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            json={"foo": "bar"},
            params=None,
            content=None,
        )

    async def test_post_with_content_length(self):
        await self.client.post(
            "/foo/bar", data={"foo": "bar"}, content_length=123
        )

        self.http_client.post.assert_awaited_once_with(
            f"{DEFAULT_GITHUB_API_URL}/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
                "Content-Length": "123",
            },
            json={"foo": "bar"},
            params=None,
            content=None,
        )

    async def test_put(self):
        await self.client.put("/foo/bar", data={"foo": "bar"})

        self.http_client.put.assert_awaited_once_with(
            f"{DEFAULT_GITHUB_API_URL}/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            json={"foo": "bar"},
            params=None,
            content=None,
        )

    async def test_put_url(self):
        await self.client.put("https://github.com/foo/bar", data={"foo": "bar"})

        self.http_client.put.assert_awaited_once_with(
            "https://github.com/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            json={"foo": "bar"},
            params=None,
            content=None,
        )

    async def test_patch(self):
        await self.client.patch("/foo/bar", data={"foo": "bar"})

        self.http_client.patch.assert_awaited_once_with(
            f"{DEFAULT_GITHUB_API_URL}/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            json={"foo": "bar"},
            params=None,
            content=None,
        )

    async def test_patch_url(self):
        await self.client.patch(
            "https://github.com/foo/bar", data={"foo": "bar"}
        )

        self.http_client.patch.assert_awaited_once_with(
            "https://github.com/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            json={"foo": "bar"},
            params=None,
            content=None,
        )

    async def test_stream(self):
        response = MagicMock()
        response.__aenter__.return_value = MagicMock()
        self.http_client.stream = MagicMock()
        self.http_client.stream.return_value = response

        async with self.client.stream("/foo/bar"):
            pass

        self.http_client.stream.assert_called_once_with(
            "GET",
            f"{DEFAULT_GITHUB_API_URL}/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            follow_redirects=True,
        )

    async def test_stream_url(self):
        response = MagicMock()
        response.__aenter__.return_value = MagicMock()
        self.http_client.stream = MagicMock()
        self.http_client.stream.return_value = response

        async with self.client.stream("https://github.com/foo/bar"):
            pass

        self.http_client.stream.assert_called_once_with(
            "GET",
            "https://github.com/foo/bar",
            headers={
                "Accept": DEFAULT_ACCEPT_HEADER,
                "Authorization": "token token",
                "X-GitHub-Api-Version": GITHUB_API_VERSION,
            },
            follow_redirects=True,
        )

    async def test_context_manager(self):
        async with self.client:
            pass

        self.http_client.__aenter__.assert_awaited_once()
        self.http_client.__aexit__.assert_awaited_once()
